#include <ctype.h>
#include <string.h>
#include <malloc.h>

#include <isode/quipu/name.h>
#include <isode/quipu/ds_search.h>
#include <isode/quipu/modify.h>
#include <isode/quipu/dua.h>

#include "bulkload.h"
#include "bulkloadglobal.h"
#include "bulkloadattrnames.h"
#include "configuration.h"
#include "filter.h"
#include "modify.h"
#include "namelist.h"
#include "processandupdate.h"
#include "processbulkandupdatedir.h"
#include "readsearch.h"
#include "subdirectory.h"
#include "util.h"

extern	struct bulkLoadConfigStruct bulkLoadConfig ;



#define	ALGONAMESEPERATOR 	'#'

#define MAXFORENAMESORINITS	5



struct 	nameComponents {

	short	titleType ; /* MALETITLEINDEX, FEMALETITLEINDEX, HUMANTITLEINDEX or NONHUMANTITLEINDEX */
	char *	titlecp ;
	
	char *	surnamecp ;

	char *	forenamesOrInitscp[ MAXFORENAMESORINITS ] ;
	short	numForenamesOrInits ;
} ;


#define NULLNAMECOMP	((struct nameComponents *) 0)


void
initialiseNameComponents( nameStructp )
struct 	nameComponents * nameStructp ;
{
	nameStructp -> titleType 	   = INVALIDTITLETYPE ;
	nameStructp -> titlecp		   = NULLCP ;
	nameStructp -> surnamecp     	   = NULLCP ;
	nameStructp -> numForenamesOrInits = 0 ;

	return ;
}

void
freeNameComponents( nameStructp )
struct 	nameComponents * nameStructp ;
{
int	index ;

	if( nameStructp )
	{
	    if( nameStructp -> titlecp )
		free( nameStructp -> titlecp ) ;

	    for( index = 0 ; index < nameStructp -> numForenamesOrInits ; index++ )
	    {
	 	if( nameStructp -> forenamesOrInitscp[ index ] )
		{
		    free( nameStructp -> forenamesOrInitscp[ index ] ) ;
		    nameStructp -> forenamesOrInitscp[ index ] = NULLCP ;
		}
	    }

	    if( nameStructp -> surnamecp )
	    {
	    	free( nameStructp -> surnamecp ) ;
		nameStructp -> surnamecp = NULLCP ;
	    }

	    initialiseNameComponents( nameStructp ) ;
	}
	
	return ;
}

char *
convNameComponentsToString( nameStructp )
struct	nameComponents * nameStructp ;
{
static	char nameBuffer[ SMALLBUFLEN ] ;
int	compNum ;

	nameBuffer[ 0 ] = '\0' ;

	if( nameStructp -> titlecp != NULLCP )
	{
	    (void)strcpy( nameBuffer, nameStructp -> titlecp ) ;
	    (void)strcat( nameBuffer, " " ) ;
	}
	
	for( compNum = 0 ; compNum < nameStructp -> numForenamesOrInits ; compNum++ )
	{
	    (void)strcat( nameBuffer, nameStructp -> forenamesOrInitscp[ compNum ] ) ;
	    (void)strcat( nameBuffer, " " ) ;
	}
	    
	if( nameStructp -> surnamecp != NULLCP )
	{
	    (void)strcat( nameBuffer, nameStructp -> surnamecp ) ;
	    (void)strcat( nameBuffer, " " ) ;
	}

	return( nameBuffer ) ;
}


/*
 * Check for a possible match between the Male, Female, Human and 
 * non-Human title types.
 */

int
possibleTitleTypesMatch( titleType1, titleType2 )
int	titleType1, titleType2 ;
{
int	possibleMatch = FALSE ;


	if( titleType1 == INVALIDTITLETYPE ||
	    titleType2 == INVALIDTITLETYPE   )
	{
	    /* .. either name does not have a title */
	    possibleMatch = TRUE ;
	}
	if( titleType1 == titleType2 )
	{
	    possibleMatch = TRUE ;
	}
	else if( titleType1 == HUMANTITLETYPE )
	{
	    if( titleType2 == MALETITLETYPE ||
	        titleType2 == FEMALETITLETYPE )
		possibleMatch = TRUE ;
	}
	else if( titleType2 == HUMANTITLETYPE )
	{
	    if( titleType1 == MALETITLETYPE ||
	        titleType1 == FEMALETITLETYPE )
		possibleMatch = TRUE ;
	}

#ifdef BLT_DEBUG
fprintf( stderr, "Possible title match between %d and %d = %d\n", titleType1, titleType2,  possibleMatch ) ;
#endif

	return( possibleMatch ) ;
}


int
matchPersonalTitle( cnCompcp, nameStructp )
char * 	cnCompcp ;
struct	nameComponents * nameStructp ;
{
int	titleType, titleIndex, titleMatched = FALSE ;

	if( nameStructp == NULLNAMECOMP )
	    return( FALSE ) ;

	for( titleType = MINTITLETYPE ; titleMatched == FALSE && titleType <= MAXTITLETYPE ; titleType++ )
	{
	    for( titleIndex = 0 ; titleMatched == FALSE && titleIndex <= bulkLoadConfig.lastUsedTitlesToken[ titleType ] ; titleIndex++ )
	    {
		if( lexequ( cnCompcp, bulkLoadConfig.titleTokenscp[ titleType ][ titleIndex ] ) == 0 )
		{
		    nameStructp -> titleType = titleType ;

		    if( nameStructp -> titlecp != NULLCP )
		    	free( nameStructp -> titlecp ) ;

		    nameStructp -> titlecp = strdup( cnCompcp ) ;
#ifdef BLT_DEBUG
fprintf( stderr, "title match for %s\n", cnCompcp ) ;
#endif
		    titleMatched = TRUE ;
		}
	    }
	}

	return( titleMatched ) ;
}

/*
 * Set up a "struct nameComponents" from the given "commonName" string..
 *
 * Note: assumes name components are delimited by single whitespaces.
 * A personal title may be present in the "commonName" itself.
 *
 * A title may be contained within the "commonName" or it
 * could be provided by a "personalTitle" component as received in "titlecp"..
 */

void
extractNameComponents( commonNamecp, titlecp, nameStructp )
char * 	commonNamecp, * titlecp ;
struct	nameComponents * nameStructp ;
{
char *	currCnCompcp, * nextCnCompcp, * basecp, * cp ;
int	compNum=0 ;	

	if( commonNamecp == NULLCP || (basecp = strdup( commonNamecp )) == NULLCP )
	{
	    nameStructp = NULLNAMECOMP ;
	    return ;
	}

	currCnCompcp = basecp ;

	/* Skip any leading {T.61} component.. */
	currCnCompcp = skipHeader( currCnCompcp, (int*)NULL ) ;

	/* Replace any full stops with a space.. */
	cp = currCnCompcp ;

	while( *cp )
	{
	    if( *cp == '.' )
	    {
		if( *(cp + 1) != '\0' )
		    *cp++ = ' ' ;
	    	else
		    *cp = '\0' ; /* Ensure string does not have a trailing space */
	    }
	    else cp++ ;
	}

	/* Replace whitespaces with a single ' ' */
	currCnCompcp = TidyString( currCnCompcp ) ;

	/* Prepare the name structure.. */
	initialiseNameComponents( nameStructp ) ;


	/* Did another RDN component have the personalTitle attribute ? */
	
	if( titlecp != NULLCP )
	    (void)matchPersonalTitle( titlecp, nameStructp ) ;

	do {

	    compNum++ ;

#ifdef BLT_DEBUG
fprintf( stderr, "Name comp %s\n", currCnCompcp ) ;
#endif
	    /* Prepare for the next name component.. */
	    nextCnCompcp = strchr( currCnCompcp, ' ' ) ;

	    if( nextCnCompcp )
	    {
	    	/* Null terminate the current component and move to the next.. */
	    	*nextCnCompcp = '\0', nextCnCompcp++ ;
	    }
	    
	    /* 
	     * The first component may be a personal title.. 
	     */
	    if( compNum > 1 || matchPersonalTitle( currCnCompcp, nameStructp ) == FALSE )
	    {
		/* ..component is not a title */
	    	
		if( nextCnCompcp == NULLCP )
	    	{
		    if( bulkLoadConfig.commonNameSearchFormat == STARSURNAMESPACEINITSTAR &&
		    	nameStructp -> surnamecp != NULLCP )
		    {
		    	/* Swiss format final forename/initial.. */
		    	if( nameStructp -> numForenamesOrInits < MAXFORENAMESORINITS )
		    	    nameStructp -> forenamesOrInitscp[ nameStructp -> numForenamesOrInits++ ] =
 		    	    	strdup( currCnCompcp ) ;
		    }
		    else
		    {
	    	    	/* We have reached the surname.. */
	    	    	nameStructp -> surnamecp = strdup( currCnCompcp );
		    }
	    	}
	    	else
	    	{
		    if( bulkLoadConfig.commonNameSearchFormat == STARSURNAMESPACEINITSTAR &&
		        nameStructp -> surnamecp == NULLCP )
		    {
		    	/* Swiss format surname.. */
	    	    	nameStructp -> surnamecp = strdup( currCnCompcp );
		    }
		    else
		    {
	    	    	/* Assume component is a forename - though could be an unknown title.. */
		    	if( nameStructp -> numForenamesOrInits < MAXFORENAMESORINITS )
		    	    nameStructp -> forenamesOrInitscp[ nameStructp -> numForenamesOrInits++ ] =
 		    	    	strdup( currCnCompcp ) ;
		    }
	    	}
	    }


	    /* Move to the next name component.. */
	    currCnCompcp = nextCnCompcp ;

	} while( currCnCompcp ) ; 

	
	free( basecp ) ;

	return ;
}

/*
 * Find common name and title components.. 
 */

int
setUpTheNameStruct( entryRdn, attrLinep, nameStructp )
RDN	entryRdn ;
struct  attributeLine * attrLinep ;
struct	nameComponents * nameStructp ;
{
int	ret_val = NOTOK ;
RDN	rdnComp ;
char *	entryRDNcommonNamecp, * titleRDNcp ;
struct  attributeLine * currALp = attrLinep ;
static	AttributeType	personalTypeat = NULLAttrT ;

	/* Find the common name RDN component of the source RDN (may be multi-valued) */
	if( (rdnComp = findAttrTypecpRdnComp( BL_COMMON_NAME, entryRdn ) ) != NULLRDN  &&
	    (entryRDNcommonNamecp = AttrV2pstr( &(rdnComp->rdn_av) )) != NULLCP )
	{
	    titleRDNcp = NULLCP ;
	
	    if( (rdnComp = findAttrTypecpRdnComp( BL_PERSONAL_TITLE, entryRdn ) ) != NULLRDN )
	    	titleRDNcp = AttrV2pstr( &(rdnComp->rdn_av) ) ;
	    else
	    {
		if( personalTypeat == NULLAttrT )
		    personalTypeat = str2AttrT( BL_PERSONAL_TITLE );
	    	
		/* Check through the attribute list for a personalTitle attribute.. */
		while( currALp && titleRDNcp == NULLCP )
		{
		    if( AttrT_cmp( currALp -> at, personalTypeat ) == 0 )
		    	titleRDNcp = AttrV2pstr( currALp -> av ) ;
		    else
		    	currALp = currALp -> nextAttrLinep ;
		}
	    }

	    /* Also check for a personalTitle component.. */
	    
	    extractNameComponents( entryRDNcommonNamecp, titleRDNcp, nameStructp ) ;

	
	    if( titleRDNcp != NULLCP )
	    	free( titleRDNcp ) ;

	    free( entryRDNcommonNamecp ) ;

	    ret_val = OK ;
	}

	return( ret_val ) ; /* OK or NOTOK */
}


/*
 * Define and action a search to find the entries with a "commonName" attribute
 * of the form INITIAL*SURNAME.
 * Return the commonName and personalTitle attributes.
 */

int
defineAndActionCnAndTitleSearch( parentDN, entryRDN, attrLinep, nameStructp, searchListpp, 
						rootedAtSubDirListDetailsp, errorDiagnosticscp )
DN	parentDN ;
RDN	entryRDN ;
struct  attributeLine * attrLinep ;
struct	nameComponents * nameStructp ;
struct 	namelist ** searchListpp ;
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
char *	errorDiagnosticscp ;
{
int	ret_val = SEARCHNOTOK, subDirListUsed ;
struct	ds_search_arg sarg;
struct	ds_search_result sresult;
struct	DSError serror ;
char 	firstInit[2] ;
struct 	namelist * attrsToBeReturnedp = NULLLIST ;

#ifdef BLT_DEBUG
struct	namelist * lp ;
#endif
	if( setUpTheNameStruct( entryRDN, attrLinep, nameStructp ) == OK )
	{
	    if( nameStructp->numForenamesOrInits <= 0 )
		(void)strcpy( errorDiagnosticscp, "The commonName component does not have a forename or initial" ) ;

	    else if( nameStructp->surnamecp == NULLCP )
		(void)strcpy( errorDiagnosticscp, "The surname component could not be processed" ) ;

	    else {

	    	/* We require the "commonName" and "personalTitle" attributes to be retrieved.. */

	    	addToList( &attrsToBeReturnedp, BL_COMMON_NAME ) ;
	    	addToList( &attrsToBeReturnedp, BL_PERSONAL_TITLE ) ;

		/* also object class and algoName since an algorithmic name attribute will need to be added.. */

	    	addToList( &attrsToBeReturnedp, BL_OBJECT_CLASS ) ;
	    	addToList( &attrsToBeReturnedp, BL_ALGO_NAME ) ;

		
		subDirListUsed = rootedAtSubDirListDetailsp -> subDirListUsed ;

		sarg = * fillMostSearchArgs(
			(subDirListUsed == SUBDIRLISTNOTUSED) ? parentDN     : rootedAtSubDirListDetailsp -> bulkSubTreeDN,
			(subDirListUsed == SUBDIRLISTNOTUSED) ? SRA_ONELEVEL : SRA_WHOLESUBTREE, attrsToBeReturnedp, FALSE ) ;


	    	/* Form a search for "First initial*Surname" */

	    	strncpy( firstInit, nameStructp->forenamesOrInitscp[ 0 ], 1 ) ;
	    	firstInit[1] = '\0' ;

#ifdef BLT_DEBUG
fprintf( stderr, "FIRST INIT [%s]\n", firstInit ) ;
#endif
	    	makeNameSearchFilter( &sarg.sra_filter, firstInit, nameStructp->surnamecp ) ;

	    	/* Carry out the search and produce a list of results.. */
	
		ret_val = RetrieveAndListAttrs( searchListpp, &sarg, &sresult, &serror ) ;

            	if (ret_val != SEARCHOK)
 		{
  	    	    logListSuccess( LIST_ERROR, "bulkDataDetails", 0) ;
 		    ds_error_free( &serror ) ;
 		}
	    	else
{
		    filterSubTreeNamelist( rootedAtSubDirListDetailsp, searchListpp ) ;
	    	    logListSuccess( LIST_OK, "bulkDataDetails", listlen( *searchListpp )) ;

#ifdef BLT_DEBUG
lp = *searchListpp ;

if( lp == NULLLIST )
	fprintf( stderr, "Nothing returned from the search\n" ) ;

while( lp != NULLLIST )
{
	fprintf( stderr, "Cn and title search %s\n", lp->name ) ;
	lp = lp -> next ;
}
#endif
}
	    	freeListpp( &attrsToBeReturnedp ) ;

		freeSearchArgs( &sarg ) ;
	    }			    
	}	
	/* else no commonName component */
	
	return( ret_val ) ; /* SEARCHNOTOK, SEARCHOK,... [bulkloadglobal.h] */
}


/*
 * ****** Note: returns a pointer to a static buffer *******
 */

char *
formAlgoName( sourceNamecp, rdncp )
char *	sourceNamecp, * rdncp ;
{
static	char algoNameBuffer[ SMALLBUFLEN ] ;

	sprintf( algoNameBuffer, "%s %c %s", sourceNamecp, ALGONAMESEPERATOR, rdncp ) ;
#ifdef BLT_DEBUG
fprintf( stderr, "Derive algoname from source %s and rdn %s\n", sourceNamecp, rdncp ) ;
#endif
	return( algoNameBuffer ) ;
}

/*
 * ****** Note: returns a pointer to a static buffer *******
 */

char *
formAlgoNameFromEntryAndSubDirListRDNs( sourceNamecp, entryRDN, subDirListRDN )
char *	sourceNamecp ;
RDN	entryRDN,
	subDirListRDN ; /* May be NULLRDN */
{
char *	algoNamecp = NULLCP, *algoNameRDNcp ;
RDN	algoNameRDN = NULLRDN ;

	if( (algoNameRDN = rdn_cpy( entryRDN )) != NULLRDN )
	{
	    if( subDirListRDN != NULLRDN )
		algoNameRDN = rdn_merge( algoNameRDN, rdn_cpy( subDirListRDN )) ;

	    if( (algoNameRDNcp = rdn2pstr( algoNameRDN )) != NULLCP )
	    {
	    	algoNamecp = formAlgoName( sourceNamecp, algoNameRDNcp ) ;
	    	free( algoNameRDNcp ) ;
	    }

	    rdn_free( algoNameRDN ) ;
	}

	return( algoNamecp ) ;
}


/*
 * Confirm that a unique key search is permitted and that the attribute list
 * contains one or more instances of the required attribute type(s).
 * If so create the necessary search filter for return.
 */
int
validateUniqueKeySearchAndFormFilter( attrLinep, uniqueKeyFilterpp )
struct  attributeLine * attrLinep ;
struct  s_filter ** uniqueKeyFilterpp ;
{
int	ret_status = NOTOK ;
struct  attributeLine * currALp ;
struct	uniqueKeyNode * currUKnodep ;
char *	cp ;

	*uniqueKeyFilterpp = (struct s_filter *)NULL ;

	if( bulkLoadConfig.uniqueKeyNodep != NULLUNIQUEKEYNODE &&
	    bulkLoadConfig.uniqueKeyNodep -> uniqueKeyAttrT != NULLAttrT ) {

	    /* Check for the required unique key attribute types.. */
	    for( currUKnodep =  bulkLoadConfig.uniqueKeyNodep ;
			currUKnodep != NULLUNIQUEKEYNODE ; currUKnodep = currUKnodep->nextUniqueKeyNodep ) {

		for( currALp = attrLinep ; currALp != NULLALP ; currALp = currALp -> nextAttrLinep ) {

		    /* If this attribute's type is a unique key attribute type we
		       only utilise the type/value pairing if the attribute/value
		       has not been marked for deletion. */

		    if( (cp = currALp->attrLinecp) != NULLCP &&
			(*cp != DELETEATTRCHAR) && (*cp != DELETEVALUECHAR) &&
			AttrT_cmp( currALp->at, currUKnodep->uniqueKeyAttrT ) == 0 ) {

			/* Append this attribute type/value to the current filter.. */
			appendEqAttrFilter( uniqueKeyFilterpp, currALp -> at, currALp -> av ) ;

			ret_status = OK ;
		    }
		}
	    }
	}

	return ret_status ;
}


/*
 * Check that the attribute list holds values for one (or more) of the
 * required unique key attribute types.
 * If so perform a search for the required attribute type value assertions.
 * If the required entry is found then insert the algorithmic name.
 *
 * Note 1 : this routine will only be called if the corresponding directory
 * entry does not have an algoname.
 *
 * Note 2 : Should two entries be returned with the same unique key attribute value
 * then a serious error is flagged.
 */

int
defineAndActionUniqueKeySearch( parentDN, searchListpp, rootedAtSubDirListDetailsp, fp, searchAlgp, errorDiagnosticscp )
DN	parentDN ;
struct 	namelist ** searchListpp ;
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
struct  s_filter * fp ;
searchAlgEnum * searchAlgp ;
char *	errorDiagnosticscp ;
{
int	ret_val = SEARCHNOTOK, subDirListUsed ;
struct	ds_search_arg sarg;
struct	ds_search_result sresult;
struct	DSError serror ;
char *	errorcp=NULLCP ;
static	struct namelist  * attrsToBeReturnedp = NULLLIST ;
struct 	namelist * searchListp, *errListp ;
int	errorCounter = 0 ; /* To avoid over running the error buffer */

	*searchAlgp = uniqueKeySearchAlg ;

	if( attrsToBeReturnedp == NULLLIST ) {
	    /* Add those attribute types that are used by addAlgoNameAndBulkObjCls() */
	    addToList( &attrsToBeReturnedp, BL_OBJECT_CLASS ) ;
	    addToList( &attrsToBeReturnedp, BL_ALGO_NAME ) ; /* Other algonames may be present */
	}

	subDirListUsed = rootedAtSubDirListDetailsp -> subDirListUsed ;

	sarg = * fillMostSearchArgs(
			(subDirListUsed == SUBDIRLISTNOTUSED) ? parentDN     : rootedAtSubDirListDetailsp -> bulkSubTreeDN,
			(subDirListUsed == SUBDIRLISTNOTUSED) ? SRA_ONELEVEL : SRA_WHOLESUBTREE, attrsToBeReturnedp, FALSE ) ;

	sarg.sra_filter = fp ;

	/* Carry out the search and produce a list of results.. */
	
	ret_val = RetrieveAndListAttrs( searchListpp, &sarg, &sresult, &serror ) ;

        if (ret_val != SEARCHOK)
	{
	    logListSuccess( LIST_ERROR, "bulkDataDetails", 0) ;

	    errorcp = error2pstr( &serror ) ; /* This will lead to a call to ds_error_free() */

	    if( errorcp != NULLCP )
	    {
	    	(void)strcpy( errorDiagnosticscp, errorcp ) ;

	    	free( errorcp ) ;
	    }
	}
	else
	{
	    filterSubTreeNamelist( rootedAtSubDirListDetailsp, searchListpp ) ;
	    logListSuccess( LIST_OK, "bulkDataDetails", listlen( *searchListpp )) ;
	}

	freeSearchArgs( &sarg ) ;


	if( ret_val == SEARCHOK && (searchListp = (*searchListpp)) != NULLLIST ) {

	    /* Confirm that only one entry has been matched.. */
	    if( searchListp->next != NULLLIST &&
		dn_cmp( searchListp->entryDN, searchListp->next->entryDN ) != 0 ) {

		(void)strcpy( errorDiagnosticscp, "SERIOUS ERROR: Unique Key attribute search matched multiple entries..." ) ;

		for( errListp = searchListp ; errListp != NULLLIST && errorCounter < 5 ;
						errListp = errListp->next, ++errorCounter ) {

		    (void)strcat( errorDiagnosticscp, "\n# " ) ;
		    (void)strcat( errorDiagnosticscp, errListp->name ) ;
		}

		freeListpp( searchListpp ) ;
		ret_val = SEARCHNOTOK ;
	    }

	}

	return( ret_val ) ; /* SEARCHNOTOK, SEARCHOK,... [bulkloadglobal.h] */
}

/*
 * Set up the algoName attribute "sourceName#RDN" search and then action it.
 *
 * If we are using a sub directory list we merely do a whole subtree search, despite
 * any limitations specified by the SubDir attributes.
 */

int
defineAndActionAlgoNameAttrSearch( sourceNamecp, parentDN, entryRDN, attrsToBeReturnedp, allAttributes, 
					searchListpp, rootedAtSubDirListDetailsp, 
					searchAlgp, errorDiagnosticscp )
char *	sourceNamecp ;
DN	parentDN ;
RDN	entryRDN ;
struct 	namelist  * attrsToBeReturnedp ;
int	allAttributes ;
struct 	namelist ** searchListpp ;
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
searchAlgEnum * searchAlgp ;
char *	errorDiagnosticscp ;
{
int	ret_val = SEARCHNOTOK, subDirListUsed ;
struct	ds_search_arg sarg;
struct	ds_search_result sresult;
struct	DSError serror ;
char *	sourceRDNalgoNamecp, * errorcp ;

	*searchAlgp = algonameSearchAlg ;

	sourceRDNalgoNamecp = formAlgoNameFromEntryAndSubDirListRDNs( sourceNamecp, entryRDN, rootedAtSubDirListDetailsp->subDirListRDN ) ;

#ifdef BLT_DEBUG
fprintf( stderr, "AlgoName Searching for %s\n", sourceRDNalgoNamecp ) ;
#endif

	subDirListUsed = rootedAtSubDirListDetailsp -> subDirListUsed ;

	sarg = * fillMostSearchArgs(
			(subDirListUsed == SUBDIRLISTNOTUSED) ? parentDN     : rootedAtSubDirListDetailsp -> bulkSubTreeDN,
			(subDirListUsed == SUBDIRLISTNOTUSED) ? SRA_ONELEVEL : SRA_WHOLESUBTREE, attrsToBeReturnedp, allAttributes ) ;

	sarg.sra_filter = orfilter() ;
	sarg.sra_filter->FUFILT = eqfilter( FILTERITEM_EQUALITY, BL_ALGO_NAME, sourceRDNalgoNamecp ) ;

	/* Carry out the search and produce a list of results.. */
	
	ret_val = RetrieveAndListAttrs( searchListpp, &sarg, &sresult, &serror ) ;

        if (ret_val != SEARCHOK)
	{
	    logListSuccess( LIST_ERROR, "bulkDataDetails", 0) ;

	    errorcp = error2pstr( &serror ) ; /* This will lead to a call to ds_error_free() */

	    if( errorcp != NULLCP )
	    {
	    	(void)strcpy( errorDiagnosticscp, errorcp ) ;

	    	free( errorcp ) ;
	    }
	}
	else
	{
	    filterSubTreeNamelist( rootedAtSubDirListDetailsp, searchListpp ) ;
	    logListSuccess( LIST_OK, "bulkDataDetails", listlen( *searchListpp )) ;
	}

	freeSearchArgs( &sarg ) ;

	return( ret_val ) ; /* SEARCHNOTOK, SEARCHOK,... [bulkloadglobal.h] */
}


/*
 * Check if the entry's search/read attribute list has an "algoName" attribute already
 * to determine whether the ADD should be "EM_ADDATTRIBUTE" or "EM_ADDVALUES".
 *
 * Also check that the "objectClass" attributes include "bulkDataEntryDetails"
 * else if another bulk data class is not already present it must be added.
 *
 * Note: implicit from the above is that the "entryListp" search results must contain any
 *	 algoname and objectClass attributes.
 */

void
addAlgoNameAndBulkObjCls( entryListp, entryDN, sourceNamecp, entryRDN, subDirListRDN )
struct 	namelist * entryListp ; /* Points to the namelist node containing the algoName/objectClass information
				   for the entry that the new algoName attribute is required      */
DN	entryDN ;
char *	sourceNamecp ;
RDN	entryRDN, subDirListRDN ;
{
Attr_Sequence	 as  ;
AV_Sequence	 avs ;
int	attrTypeValueFoundCnt = 0, algoNameTypeFound = FALSE, bulkLoadObjClsFound = FALSE, em_type ;
char *	errorMsgcp ;
static	AttributeType  algoNameat = NULLAttrT, objClsat = NULLAttrT ;
static	AttributeValue dataEntryObjClsav = NULLAttrV, dataSourceObjClsav = NULLAttrV, dataSourceEntryObjClsav = NULLAttrV, bulkSubDirListObjClsav = NULLAttrV ;
AttributeValue algoNameav ;
char *	avcp ;
struct entrymod   *mea_changesp = NULLMOD, *base_changesp = NULLMOD ;

	if( entryListp == NULLLIST )
	    return ;
	
	if( algoNameat == NULLAttrT )
	{
	    algoNameat = str2AttrT( BL_ALGO_NAME ) ;
	}

	if( objClsat == NULLAttrT )
	{
	    objClsat = str2AttrT( BL_OBJECT_CLASS ) ;
	}

	if( dataEntryObjClsav == NULLAttrV )
	{
	    dataEntryObjClsav = str2AttrV( BL_BULK_DATA_ENTRY_DETAILS, objClsat->oa_syntax ) ;
	}

	if( dataSourceObjClsav == NULLAttrV )
	{
	    dataSourceObjClsav = str2AttrV( BL_BULK_DATA_SOURCE_DETAILS, objClsat->oa_syntax ) ;
	}

	if( dataSourceEntryObjClsav == NULLAttrV )
	{
	    dataSourceEntryObjClsav = str2AttrV( BL_BULK_DATA_SOURCE_ENTRY, objClsat->oa_syntax ) ;
	}

	if( bulkSubDirListObjClsav == NULLAttrV )
	{
	    bulkSubDirListObjClsav = str2AttrV( BL_BULK_SUB_DIR_LIST_CLASS, objClsat->oa_syntax ) ;
	}

	/* For the objectClass and algoName attribute sequences.. */
	for( as = entryListp -> ats ;  attrTypeValueFoundCnt < 2 && as != NULLATTR ; as = as -> attr_link )
	{
	    if( AttrT_cmp( as->attr_type, algoNameat ) == 0 )
	    {
		algoNameTypeFound = TRUE ;
		attrTypeValueFoundCnt++ ;
	    }
	    else if( AttrT_cmp( as->attr_type, objClsat ) == 0 )
	    {
		/* Run through the "objectClass" values */
		for( avs = as -> attr_value ; bulkLoadObjClsFound == FALSE && avs != NULLAV ; avs = avs -> avseq_next )
		{
		    if( AttrV_cmp( &(avs->avseq_av), dataSourceEntryObjClsav ) == 0 ||
			AttrV_cmp( &(avs->avseq_av), dataSourceObjClsav )      == 0 ||
			AttrV_cmp( &(avs->avseq_av), dataEntryObjClsav )       == 0 ||
			AttrV_cmp( &(avs->avseq_av), bulkSubDirListObjClsav )  == 0   )
		    {
			bulkLoadObjClsFound = TRUE ;
		    }
		}
		
		attrTypeValueFoundCnt++ ;
	    }
	}

	if( algoNameTypeFound == TRUE )
	    em_type = EM_ADDVALUES ;
	else
	    em_type = EM_ADDATTRIBUTE ;

	avcp = formAlgoNameFromEntryAndSubDirListRDNs( sourceNamecp, entryRDN, subDirListRDN ) ;

	algoNameav = str2AttrV( avcp, algoNameat -> oa_syntax ) ;

	/* Must add the new algoName value.. */

	mea_changesp = create_attr_mod( em_type, algoNameat, algoNameav ) ;
	append_attr_mod( &base_changesp, mea_changesp ) ;
	
	AttrV_free( algoNameav ) ;

	if( bulkLoadObjClsFound == FALSE )
	{
	    /* Must add the bulk load entry object class.. */
	    mea_changesp = create_attr_mod( EM_ADDVALUES, objClsat, dataEntryObjClsav ) ;
	    append_attr_mod( &base_changesp, mea_changesp ) ;
	}


	if( applyModificationsToDNentry( entryDN, base_changesp, &errorMsgcp ) == NOTOK )
{
#ifdef BLT_DEBUG
fprintf( stderr, "Algo name addition failed %s\n", errorMsgcp ) ;
#endif
	    free( errorMsgcp ) ;
}

	/* Return storage.. */
	free_attr_mod_list( base_changesp );
	
	return ;
}


/*
 * Add an algoName attribute to the end of the current attribute list.
 * If a sub-directory list is used the algoname is specific to it.
 */

void
appendAlgoNameAttributeToAttrSeq( entryRDN, subDirListRDN, sourceNamecp, as )
RDN	entryRDN, subDirListRDN ;
char *	sourceNamecp ;
Attr_Sequence	as   ;
{
char *	algorithmicNamecp ;

	if( (algorithmicNamecp = formAlgoNameFromEntryAndSubDirListRDNs( sourceNamecp, entryRDN, subDirListRDN )) != NULLCP )
	    appendAttributeToAttrSeq( BL_ALGO_NAME, algorithmicNamecp, as ) ;

	return ;
}


/*
 * Perform an algorithmic name replacement on a directory entry.
 */

void	replaceAlgoName( entryDN, newrdn, oldrdn, subDirListRDN, sourceNamecp )
DN	entryDN ;
RDN	newrdn, oldrdn, subDirListRDN ;
char *	sourceNamecp ;
{
char *	algorithmicNamecp ;
AttributeType	algoat = NULLAttrT ;
AttributeValue	algoav = NULLAttrV ;
char *	errorMsgcp ;
int	errorClass, errorType ;

#ifdef BLT_DEBUG
char *	cp ;
cp = dn2pstr( entryDN ) ;
fprintf( stderr, "entryDN = %s\n", cp ) ;
free( cp ) ;
cp = rdn2pstr( newrdn ) ;
fprintf( stderr, "newrdn = %s\n", cp ) ;
free( cp ) ;
cp = rdn2pstr( oldrdn ) ;
fprintf( stderr, "oldrdn = %s\n", cp ) ;
free( cp ) ;
fprintf( stderr, "source name = %s\n", sourceNamecp ) ;
#endif
	if( (algoat = str2AttrT( BL_ALGO_NAME )) != NULLAttrT )
	{
	    /* Perform the add algoname attribute.. */

	    algorithmicNamecp = formAlgoNameFromEntryAndSubDirListRDNs( sourceNamecp, newrdn, subDirListRDN ) ;

	    if( (algoav = str2AttrV( algorithmicNamecp, algoat -> oa_syntax )) != NULLAttrV )
	    {
	    	/* Add the new algoName value.. */
	   	if( modifyAttribute( EM_ADDVALUES, entryDN, algoat, algoav, &errorMsgcp, &errorClass, &errorType ) == NOTOK )
	    	{
	    	    free( errorMsgcp ) ;

		    /* May need to add the attribute instead.. */
		    if( errorClass == DSE_ATTRIBUTEERROR && errorType == DSE_AT_NOSUCHATTRIBUTE )
		    {
			/* Add the new algoName attribute */
	    	    	if( modifyAttribute( EM_ADDATTRIBUTE, entryDN, algoat, algoav, &errorMsgcp, &errorClass, &errorType ) == NOTOK )
	    	    	{
			    free( errorMsgcp ) ;
			}
		    }
	 	}
		    
		AttrV_free( algoav ) ;
	    }

	    /* Perform the delete algoname attribute.. */
	
    	    algorithmicNamecp = formAlgoNameFromEntryAndSubDirListRDNs( sourceNamecp, oldrdn, subDirListRDN ) ;

	    if( (algoav = str2AttrV( algorithmicNamecp, algoat -> oa_syntax )) != NULLAttrV )
	    {
	    	/* Add the new algoName value.. */
	   	if( modifyAttribute( EM_REMOVEVALUES, entryDN, algoat, algoav, &errorMsgcp, &errorClass, &errorType ) == NOTOK )
	    	{
	    	    free( errorMsgcp ) ;

		    /* Since we have added one or more values we do not require a REMOVE ATTRIBUTE here */
	 	}

		AttrV_free( algoav ) ;
	    }

	    AttrT_free( algoat ) ;
	}

	return ;
}

/* Modified by J.C. 20/04/94 to permit matches where there are different numbers
 * of forename components but there is a match as far as they go.
 * This will only be used if
 * bulkLoadConfig.updatesToExistingEntriesIfUnique == TRUE
 */
int
exactForenameMatch( nameStruct1p, nameStruct2p, liberalMatchCriterion )
struct	nameComponents	* nameStruct1p, * nameStruct2p ;
int	liberalMatchCriterion ;
{
int	exactNameMatch = TRUE, compareNforenamesOrInits = 0, compNum ;

 	/* compareNforenamesOrInits = minimum numForenamesOrInits over both nameStructs */

 	compareNforenamesOrInits = nameStruct1p->numForenamesOrInits ;
 	    
 	if (compareNforenamesOrInits > nameStruct2p->numForenamesOrInits)
 		compareNforenamesOrInits =  nameStruct2p->numForenamesOrInits ;

	/* Check for a mismatch in initials/forenames.. */

	if( (nameStruct1p->numForenamesOrInits != nameStruct2p->numForenamesOrInits ) &&
	    ( liberalMatchCriterion == FALSE ) ) {

	    exactNameMatch = FALSE ;

	} else {
	
	    for( compNum = 0 ; compNum < compareNforenamesOrInits ; compNum++ )
	    {
	    	if( nameStruct1p->forenamesOrInitscp[ compNum ][ 1 ] == '\0' ||
		    nameStruct2p->forenamesOrInitscp[ compNum ][ 1 ] == '\0'   ) {

		    /* One of the forenames is an initial */
		    if( strncasecmp( nameStruct1p->forenamesOrInitscp[ compNum ],
				     nameStruct2p->forenamesOrInitscp[ compNum ], 1 ) != 0 ) {
			exactNameMatch = FALSE ;
			break ;
		    }
		}
		else
		{
		    /* Neither of the forenames is an initial */
		    if( strcasecmp( nameStruct1p->forenamesOrInitscp[ compNum ],
				    nameStruct2p->forenamesOrInitscp[ compNum ] ) != 0 ) {
			exactNameMatch = FALSE ;
			break ;
		    }
		}
	    }
	}

	return( exactNameMatch ) ;
}

/*
 * Can we state that the two names are different,
 * e.g. one is a "Mr" the other a "Miss" or
 * one has first initials "A B" whilst the other
 * has "A C".
 */

int
possibleForenameMatch( nameStruct1p, nameStruct2p )
struct	nameComponents	* nameStruct1p, * nameStruct2p ;
{
int	possibleNameMatch = TRUE, compareNforenamesOrInits = 0, compNum ;

	if( possibleTitleTypesMatch( nameStruct1p->titleType, nameStruct2p->titleType ) == FALSE )
	    possibleNameMatch = FALSE ;
	else
	{
	    /* Check for a mismatch in initials/forenames.. */

	    compareNforenamesOrInits = (nameStruct1p->numForenamesOrInits <
	    				nameStruct2p->numForenamesOrInits  ) ?
	 	    		       nameStruct1p->numForenamesOrInits  : nameStruct2p->numForenamesOrInits ;

	    for( compNum = 0 ; compNum < compareNforenamesOrInits ; compNum++ )
	    {
	    	if( nameStruct1p->forenamesOrInitscp[ compNum ][ 1 ] == '\0' ||
		    nameStruct2p->forenamesOrInitscp[ compNum ][ 1 ] == '\0'   )
		{
		    /* One of the forenames is an initial */
		    if( strncasecmp( nameStruct1p->forenamesOrInitscp[ compNum ],
				 nameStruct2p->forenamesOrInitscp[ compNum ], 1 ) != 0 )
		    {
			possibleNameMatch = FALSE ;
			break ;
		    }
		}
		else
		{
		    /* Neither of the forenames is an initial */
		    if( strcasecmp( nameStruct1p->forenamesOrInitscp[ compNum ],
				 nameStruct2p->forenamesOrInitscp[ compNum ] ) != 0 )
		    {
			possibleNameMatch = FALSE ;
			break ;
		    }
		}
	    }
	}

	return( possibleNameMatch ) ;
}


/*
 * Check through the common name attributes that are to be added
 * when the entry is created comparing them with the current
 * commonName from an existing entry RDN to see if we can
 * disambiguate the two entries..
 */

int
diffInitOrForeNames( createEntryAttrLinep, searchNameStructp )
struct	attributeLine 	* createEntryAttrLinep ;
struct	nameComponents	* searchNameStructp    ; /* from a commonName sttribute */
{
int	ret_val = FALSE ;
struct	attributeLine * attrLinep = createEntryAttrLinep ;
char *	namecp ;
struct 	nameComponents	createNameStruct, * createNameStructp = &createNameStruct ;

	/* Find the next commonName attribute to be added.. */

	while( attrLinep && ret_val == FALSE )
	{
	    if( strcmp( attr2name( attrLinep -> at, OIDPART ), BL_COMMON_NAME ) == 0 )
	    {
	    	namecp = AttrV2pstr( attrLinep -> av ) ;

		if( namecp != NULLCP )
		{
		    extractNameComponents( namecp, NULLCP, createNameStructp ) ;

#ifdef BLT_DEBUG
fprintf( stderr, "Comparing %s\n", convNameComponentsToString( createNameStructp ) ) ;
fprintf( stderr, "to  %s\n", convNameComponentsToString( searchNameStructp ) ) ;
#endif
		    if( possibleForenameMatch( createNameStructp, searchNameStructp ) == FALSE )
			ret_val = TRUE ;		    

		    freeNameComponents( createNameStructp ) ;
		    free( namecp ) ;
		}
	    }

	    attrLinep = attrLinep -> nextAttrLinep ;
	}	
#ifdef BLT_DEBUG
fprintf( stderr, "diff result %d\n", ret_val ) ;
#endif
	return( ret_val ) ;
}

/*
 * Since the data source is providing data for entries that should already exist
 * look for an "exact match" with an existing entry. An "exact match" is defined as :-
 *
 *	"Single occurrance of a match with an existing entry which has the same initials
 *	 and surname. If the personal title is available then there must be at least
 *	 an ambiguous match."
 *
 * NOTE: the procedure will only be called if updates for existing entries are being processed.
 */

 /* Altered by J.C. 20/04/94 to allow for more liberal matching, too */
 /* Also added OFD parameter, to allow reporting to .done file when liberal matching criterion is used */

int
unambiguousUpdateDNresolution( searchListp, updateNameStructp, sourceNamecp, resolvedEntryDNp, entryRDN, subDirListRDN, OFD )
struct	namelist           * searchListp ; /* Results of a *INITIAL* SURNAME* search */
struct	nameComponents	   * updateNameStructp ;
char  * sourceNamecp ;
DN    * resolvedEntryDNp   ; /* Returned on success */
RDN	entryRDN, subDirListRDN ;
struct  outputFileDetails OFD[] ;
{
int 	ret_val = OK ;
struct	namelist * entryListp, * savedEntryListp = NULLLIST ;
Attr_Sequence	as  ;
AV_Sequence	avs ;
struct	nameComponents 	searchNameStruct, *searchNameStructp = &searchNameStruct ;
int	matchCount=0, entryLoopExit = FALSE, foreNameMatch = FALSE, possTitlesMatch = FALSE ;
int	liberalForenameMatch = FALSE;
int	liberalCount = 0;
int 	liberalMatchCriterion = FALSE ;
RDN	searchRDN, commonNameRDNcomp = NULLRDN, personalTitleRDNcomp = NULLRDN ;
char *	searchNamecp, * searchTitlecp, * commonNamecp, * titlecp ;

	liberalMatchCriterion = FALSE;
	if (bulkLoadConfig.updatesToExistingEntriesIfUnique == TRUE)
	    liberalMatchCriterion = TRUE;

	for( entryListp = searchListp ; entryLoopExit == FALSE && entryListp != NULLLIST ;
									entryListp = entryListp -> next )
	{
	    /* Run though the searchList entries */
	    
	    if( (searchRDN = findLastRDN( entryListp -> entryDN )) == NULLRDN )
	    {
	    	ret_val = NOTOK ; entryLoopExit = TRUE ;
	    }
	    else
	    {
		searchNamecp = searchTitlecp = NULLCP ;
	    
	    	/* Find the commonName RDN component for this entry.. */
	    	if( (commonNameRDNcomp = findAttrTypecpRdnComp( BL_COMMON_NAME, searchRDN ) ) != NULLRDN )
		{
		    searchNamecp = AttrV2pstr( &(commonNameRDNcomp -> rdn_av) ) ;

		    if( (personalTitleRDNcomp = findAttrTypecpRdnComp( BL_PERSONAL_TITLE, searchRDN )) != NULLRDN )
		    	searchTitlecp = AttrV2pstr( &(personalTitleRDNcomp->rdn_av) ) ;
		
	    	    extractNameComponents( searchNamecp, searchTitlecp, searchNameStructp ) ;

		    free( searchNamecp ) ;
	    	    
		    if( searchTitlecp )
	    	    	free( searchTitlecp ) ;

	    	    if( strcasecmp( updateNameStructp->surnamecp, searchNameStructp->surnamecp ) == 0 )
	    	    {
		    	/* ..same surname */

			foreNameMatch = exactForenameMatch( updateNameStructp, searchNameStructp, FALSE ) ;

			if (liberalMatchCriterion == TRUE) {
				liberalForenameMatch = exactForenameMatch( updateNameStructp, searchNameStructp, TRUE );
			}
			
			possTitlesMatch = possibleTitleTypesMatch( updateNameStructp->titleType,
 								   searchNameStructp->titleType ) ;

			freeNameComponents( searchNameStructp ) ;

			/* ..Note: will return TRUE unless both names have titles and that
			   these titles definitely do not match e.g. Mr to Mrs, Dr to Daemon */

			/* Run through the attribute types.. */
		    	for( as = entryListp->ats ; (foreNameMatch == FALSE || possTitlesMatch == TRUE) && as != NULLATTR ; as = as->attr_link )
	    		{
			    if( strcmp( attr2name( as->attr_type, OIDPART ), BL_COMMON_NAME ) == 0 )
			    {
			    	/* Run through the attribute values for this attribute type.. */
				for( avs = as->attr_value ; foreNameMatch == FALSE && avs != NULLAV ; avs = avs->avseq_next )
			        {
				    if( (commonNamecp = AttrV2pstr( &(avs->avseq_av) ) ) != NULLCP )
				    {
					extractNameComponents( commonNamecp, NULLCP, searchNameStructp ) ;

					foreNameMatch = exactForenameMatch( updateNameStructp, searchNameStructp, FALSE ) ;
			
					if (liberalMatchCriterion == TRUE) {
					    liberalForenameMatch = exactForenameMatch( updateNameStructp, searchNameStructp, TRUE);
					}

					freeNameComponents( searchNameStructp ) ;
				    	free( commonNamecp ) ;
			    	    }		    
			    	}
			    }
			    else if( strcmp( attr2name( as->attr_type, OIDPART ), BL_PERSONAL_TITLE ) == 0 )
			    {
			    	/* Run through the attribute values for this attribute type.. */
			    	for( avs = as->attr_value ; possTitlesMatch == TRUE && avs != NULLAV ; avs = avs->avseq_next )
			    	{
				    if( (titlecp = AttrV2pstr( &(avs->avseq_av) )) != NULLCP )
				    {
				    	(void)matchPersonalTitle( titlecp, searchNameStructp ) ;
			    
				    	possTitlesMatch = possibleTitleTypesMatch( updateNameStructp->titleType, searchNameStructp->titleType ) ;

					freeNameComponents( searchNameStructp ) ;
					free( titlecp ) ;
				    }
			    	}
			    }		
		    	}

			if( possTitlesMatch == TRUE )
			{
			    if( foreNameMatch == TRUE ) {
				matchCount++ ;
				savedEntryListp = entryListp ;
			    }
			    
			    if( (liberalMatchCriterion == TRUE) &&
				(liberalForenameMatch == TRUE )   ) {

				liberalCount++ ;

				if( foreNameMatch == FALSE )
				{
				    savedEntryListp = entryListp ;
				}
			    }
			}
		    }
	    	}
	    }
	}


	if( matchCount == 1 ) {

	    *resolvedEntryDNp = dn_cpy( savedEntryListp -> entryDN ) ;
	
	    if( bulkLoadConfig.disableBulkAttributeUsage == FALSE &&
	    	savedEntryListp != NULLLIST && *resolvedEntryDNp != NULLDN )
		addAlgoNameAndBulkObjCls( savedEntryListp, *resolvedEntryDNp, sourceNamecp, entryRDN, subDirListRDN ) ;

	    ret_val = OK ;

	}  else if ( liberalCount == 1) {
	
	    *resolvedEntryDNp = dn_cpy( savedEntryListp -> entryDN ) ;
	
	    /* Report this update in DONE file */

	    fprintf( OFD[ OUTPUTDONEFILE ].outFilep, "\n# Matched by surname and different numbers of forenames with DN : \n#  %s \n", savedEntryListp -> name ) ;

	    if( bulkLoadConfig.disableBulkAttributeUsage == FALSE &&
	    	savedEntryListp != NULLLIST && *resolvedEntryDNp != NULLDN )
		addAlgoNameAndBulkObjCls( savedEntryListp, *resolvedEntryDNp, sourceNamecp, entryRDN, subDirListRDN ) ;

	    ret_val = OK ;

	} else {

	    ret_val = NOTOK ;	
	}

	return( ret_val ) ; /* OK if a single exact match found with existing entry else NOTOK */
}


/* 
 * We wish to show that the entry to be created differs unambiguously from
 * all existing entries.
 *
 * Note 1 : a READ will already have failed to find an existing entry based on the RDN.
 *
 * Note 2 : If this procedure is being called to check for ambiguous entry matches
 *	    to an RDN used in a modifyRDN operation then an ambiguous match with the
 *	    entry having its RDN modified can be ignored.
 *
 * This procedure must take into account alternative "commonNames"
 * and "personalTitles".
 *
 * Case      1 :To permit the creation of Paul Barker then no entry with 
 *		RDN P Barker can exist, or if it does then it must have disambiguating 
 *		alternative commonNames or personalTitles.
 *
 *           2 :To permit the creation of Pierre Barker even if Paul Barker 
 *		already exists.
 *
 *	     3 :To permit the creation of Pat Jones with an alternative commonName
 *		of Pat Sarah Jones when an entry with an RDN of P A Jones already
 *		exists.
 *
 *	     4 :To permit the creation of P A B Barker even if P B Barker exists, but
 *		not if P A Barker exists with no disamibiguating attributes.
 */

int
unambiguousCreateEntryDNresolution( currRAAp, searchListp, createNameStructp, modRDNUpdate, diagnostics1cp, diagnostics2cp )
struct	rootedAtAttributes * currRAAp ;
struct	namelist           * searchListp ; /* Results of a *INITIAL* SURNAME* search */
struct	nameComponents     * createNameStructp ;
int	modRDNUpdate ;
char *	diagnostics1cp,    * diagnostics2cp ;
{
int 	ret_val = OK ;
struct	namelist * entryListp ;
Attr_Sequence	as  ;
AV_Sequence	avs ;
struct	nameComponents 	searchNameStruct, *searchNameStructp = &searchNameStruct ;
int	matchCouldNotBeDisambiguatedFromAnEntry = FALSE, entryLoopExit = FALSE, attrLoopsExit ;
RDN	searchRDN, commonNameRDNcomp = NULLRDN, personalTitleRDNcomp = NULLRDN, entryListRDN,
	uidRDNcomp, searchUidRDNcomp ;
char *	searchNamecp, * searchTitlecp, * commonNamecp, * titlecp ;
static	AttributeType uidat = NULLAttrT ;

	if( uidat == NULLAttrT )
	    uidat =  str2AttrT( BL_UID ) ;
	
	uidRDNcomp = findAttrTypeRdnComp( uidat, currRAAp -> rdn ) ;


	if( createNameStructp -> numForenamesOrInits <= 0 )
	{
	    (void)strcpy( diagnostics1cp, "The commonName component does not have a forename or initial" ) ;
 	    ret_val = NOTOK ;

	} else for( entryListp = searchListp ; entryLoopExit == FALSE && entryListp != NULLLIST ;
									entryListp = entryListp -> next )
	{
	    if( modRDNUpdate == TRUE )
	    {
	    	/* We are ensuring that the modified RDN does not have even an ambiguous match with
		   any OTHER entry (although it can ambiguously match its own entry's RDN
		   e.g. "commonName = Paul Barker % personalTitle = Mr"  could be modified to
		        "commonName = Paul Barker % personalTitle = Dr")			    */		   

		if( (entryListRDN = findLastRDN( entryListp -> entryDN )) != NULLRDN &&
		    (rdn_cmp( currRAAp -> rdn, entryListRDN ) == 0 )                 )
	    	    continue ; /* Skip the entry having its RDN modified */
	    }
	
	    /* Run though the searchList entries.. */

#ifdef BLT_DEBUG
fprintf( stderr, "Entry name %s \n", entryListp -> name ) ;
#endif
	    matchCouldNotBeDisambiguatedFromAnEntry = FALSE ;

	    if( (searchRDN = findLastRDN( entryListp -> entryDN )) != NULLRDN )
	    {
		searchNamecp = searchTitlecp = NULLCP ;
	    
	    	/* Find the commonName RDN component for this entry.. */
	    	if( (commonNameRDNcomp = findAttrTypecpRdnComp( BL_COMMON_NAME, searchRDN ) ) != NULLRDN )
		{
		    searchNamecp = AttrV2pstr( &(commonNameRDNcomp -> rdn_av) ) ;

		    if( (personalTitleRDNcomp = findAttrTypecpRdnComp( BL_PERSONAL_TITLE, searchRDN )) != NULLRDN )
		    	searchTitlecp = AttrV2pstr( &(personalTitleRDNcomp->rdn_av) ) ;
		
	    	    extractNameComponents( searchNamecp, searchTitlecp, searchNameStructp ) ;

		    free( searchNamecp ) ;
	    	    
		    if( searchTitlecp )
	    	    	free( searchTitlecp ) ;


	    	    if( strcasecmp( createNameStructp->surnamecp, searchNameStructp->surnamecp) == 0 )
	    	    {
		    	/* ..same surname */

			if( uidRDNcomp != NULLRDN &&
			    (searchUidRDNcomp = findAttrTypeRdnComp( uidat, searchRDN )) != NULLRDN &&
			    rdn_cmp( uidRDNcomp, searchUidRDNcomp ) != 0 )
			{
			    /* The entries have been disambiguated by their RDN user id components */
			}
			else if( (exactForenameMatch( createNameStructp, searchNameStructp, bulkLoadConfig.updatesToExistingEntriesIfUnique ) == TRUE) &&
			         (possibleTitleTypesMatch( createNameStructp->titleType, searchNameStructp->titleType ) == TRUE) )
			{
			    /* The RDN forenames/initials match */

			    (void)strcpy( diagnostics1cp, "Full forename or initials match with the entry with DN" ) ;
			    (void)strcpy( diagnostics2cp, entryListp -> name ) ;

			    ret_val = NOTOK ; entryLoopExit = TRUE ;

			} 
			else if( possibleForenameMatch( createNameStructp, searchNameStructp ) == TRUE )
			{
			    /* titles and forenames do not disambiguate the create and search 
			       entries so we must check through the alternative attributes.. */
#ifdef BLT_DEBUG
fprintf( stderr, "Ambiguity over title match between %d and %d\n", createNameStructp->titleType, searchNameStructp->titleType ) ;
#endif
			    matchCouldNotBeDisambiguatedFromAnEntry = TRUE ;
		    	}	            
		    }

		    freeNameComponents( searchNameStructp ) ;
	    	}
	    }


	    if( matchCouldNotBeDisambiguatedFromAnEntry == TRUE )
	    {
		attrLoopsExit = FALSE ;
	    
	    	/* Run through the attribute types.. */
	    	for( as = entryListp->ats ; attrLoopsExit == FALSE && as != NULLATTR ; as = as->attr_link )
	    	{
		    if( strcmp( attr2name( as->attr_type, OIDPART ), BL_COMMON_NAME ) == 0 )
		    {
		    	/* Run through the attribute values for this attribute type.. */
			for( avs = as->attr_value ; attrLoopsExit == FALSE && avs != NULLAV ; avs = avs->avseq_next )
		        {
			    if( (commonNamecp = AttrV2pstr( &(avs->avseq_av) ) ) != NULLCP )
			    {
			    	extractNameComponents( commonNamecp, NULLCP, searchNameStructp ) ;

				/* Attempt to match the current search entry common name, first against
				   the common name RDN component of the entry to be created, else
				   try to match any of the common name attributes that will be added
				   upon entry creation. */

				if( possibleForenameMatch( createNameStructp, searchNameStructp ) == FALSE ||
				    diffInitOrForeNames( currRAAp->attrLinep, searchNameStructp ) == TRUE    )
				{
#ifdef BLT_DEBUG
fprintf( stderr, "Diff in names %d\n", ret_val) ;
#endif
			    	    /* Create entry RDN does not match this entry */
			    	    attrLoopsExit = TRUE ;
				}			    	
				
				freeNameComponents( searchNameStructp ) ;
			    	free( commonNamecp ) ;
		    	    }		    
		    	}
		    }
		    else if( strcmp( attr2name( as->attr_type, OIDPART ), BL_PERSONAL_TITLE ) == 0 )
		    {
		    	/* Run through the attribute values for this attribute type.. */
		    	for( avs = as->attr_value ; attrLoopsExit == FALSE && avs != NULLAV ; avs = avs->avseq_next )
		    	{
			    titlecp = AttrV2pstr( &(avs->avseq_av) ) ;

			    (void)matchPersonalTitle( titlecp, searchNameStructp ) ;			    
			    
			    if( possibleTitleTypesMatch( createNameStructp->titleType, searchNameStructp->titleType ) == FALSE )
			    {
			    	/* Create entry title does not match this entry */
			    	attrLoopsExit = TRUE ;
			    }

			    freeNameComponents( searchNameStructp ) ;
			    free( titlecp ) ;
		    	}
		    }		
	    	}

		if( attrLoopsExit == FALSE )
		{
		    /* We can not state that the entry to be created 
		       differs unambiguously from this search entry */

		    (void)strcpy( diagnostics1cp, "Could not be disambiguated from the entry with DN" ) ;
		    (void)strcpy( diagnostics2cp, entryListp -> name ) ;

		    ret_val = NOTOK ; entryLoopExit = TRUE ;
		}
	    }
	}

#ifdef BLT_DEBUG
fprintf( stderr, "Leaving unambiguous create with retval %d\ndiag1 %s\ndiag2 %s\n", ret_val, diagnostics1cp, diagnostics2cp ) ;
#endif
	return( ret_val ) ; /* OK if no match with existing entry else NOTOK */
}


/*
 * Called only in interactive mode.
 */

DN
userInputEntryOrParentDNresolution( parentDN, purportedDN, rootedAtSubDirListDetailsp,
		purportedRDN, sourceNamecp, parentDNreq, createEntry, diagnostics1cp, diagnostics2cp )
DN	parentDN, purportedDN ;
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
RDN	purportedRDN ;
char *	sourceNamecp ;
int	createEntry, /* If FALSE then updating an existing entry */
	parentDNreq ;
char *	diagnostics1cp, * diagnostics2cp ; /* Pointers to ERRORBUFSIZE buffers */
{
DN	resolvedEntryDN = NULLDN, resolvedParentDN = NULLDN ;
int	DNindex = 0, selection = 0, maxSelection = 0, subDirListUsed ;
DN	subDirDNext = NULLDN ;
char *	DNcp, * RDNcp ;
struct	namelist * readListp = NULLLIST ;

	subDirListUsed = rootedAtSubDirListDetailsp->subDirListUsed ;

	for( ; ; )
	{
	    resolvedEntryDN = NULLDN ; selection = 0 ;

	    if( subDirListUsed != SUBDIRLISTNOTUSED )
	    {
	    	DNcp = dn2pstr( purportedDN ) ;
	    	resetprint( "\nSource entry name =\n%s\n", DNcp ) ;
	    	free( DNcp );

	    	if( subDirListUsed == SUBDIRLISTALLSUBORDSUSED )
		{
		    /* The subordinates RDNs may not have been set up already.. */
		    if( rootedAtSubDirListDetailsp->bulkSubTreeDNextensions == ((DN *)NULL ))
		    {
		    	/* Retrieve a list of the sub-ordinates.. */
			setUpTheListAllSubordinateDNextensions( rootedAtSubDirListDetailsp, diagnostics1cp ) ;
		    }
	    	}

		if( rootedAtSubDirListDetailsp->bulkSubTreeDNextensions == ((DN *)NULL ) )
		{
		    /* Check that the diagnostic has not been defined in setUpTheListAllSubordinateDNextensions()" */
		    if( *diagnostics1cp != '\0' )
	    		(void)strcpy( diagnostics1cp, "There are no DNs listed for the sub-tree extensions list" ) ;

	    	    break ;
		}

		/* List the sub-directory DN extensions list requesting operator selection.. */
		resetprint( "\n" );

		if( (DNcp = dn2pstr( rootedAtSubDirListDetailsp->bulkSubTreeDN )) != NULLCP )
		{
		     pageprint( "The sub-directory list sub-tree name is\n%s\n", DNcp ) ;
		     free( DNcp ) ;
		}

	    	if( createEntry == TRUE )
		{
		    RDNcp = rdn2pstr( purportedRDN ) ;
		    pageprint( "What should be the next component when creating the entry\n%s\n", (RDNcp != NULLCP) ? RDNcp : "" ) ;

		    if( RDNcp != NULLCP )
		    	free( RDNcp ) ;
		}
		else
		    pageprint( "What should be the next component of an existing entry that is to be updated\n" ) ;

		for( DNindex = 0, subDirDNext = *(rootedAtSubDirListDetailsp->bulkSubTreeDNextensions + DNindex) ;
		     subDirDNext != NULLDN && selection <= 0 ;
		     ++DNindex, subDirDNext = *(rootedAtSubDirListDetailsp->bulkSubTreeDNextensions + DNindex) )
		{
		    if( (DNcp = dn2pstr( subDirDNext )) != NULLCP )
	    	    {
	    		pageprint( "%d) %s\n", DNindex+1, DNcp ) ;
	    		free( DNcp ) ;

	    		/* Retrieve the selection entered at the pager prompt */
    	    		selection = getpnum() ;
		    }
	    	}

	    	if( selection <= 0 )
    		{
		    maxSelection = DNindex+1 ;

    		    /* Allow for the possibility that it is none of these.. */
	    	    pageprint( "%d) NONE OF THE ABOVE\n", maxSelection );

		    selection = enterSelection( 1, maxSelection, NULLCP ) ;
		}

		if( selection == NONE_OF_THE_SELECTION || selection == NO_SELECTION )
		{
		    (void)strcpy( diagnostics1cp, "You have rejected all the possible sub-directories" ) ;
		    break ;
		}
		else if( selection <= 0 || selection > maxSelection )
		{
		    (void)strcpy( diagnostics1cp, "Internal error when selecting a sub-directory" ) ;
		    break ;
		}

		resolvedParentDN = dn_cpy( rootedAtSubDirListDetailsp->bulkSubTreeDN ) ;

	    	subDirDNext = *(rootedAtSubDirListDetailsp->bulkSubTreeDNextensions + selection - 1) ;

	    	if( subDirDNext != NULLDN )
	    	    dn_append( resolvedParentDN, dn_cpy( subDirDNext ) ) ;
	    }
	    else
	    {
	    	/* ..We are not dealing with a sub-directory search list */
		resolvedParentDN = dn_cpy( parentDN ) ;
	    }


	    if( parentDNreq == TRUE )
	    	return( resolvedParentDN ) ;


	    if( createEntry == FALSE )
	    {
	    	/* .. update of an existing Directory enrty */

		if( (resolvedEntryDN = enterUpdateEntryRDNandFormDN( resolvedParentDN,
						purportedDN, &readListp )) == NULLDN )
		{
		    if( subDirListUsed == SUBDIRLISTNOTUSED )
		    {
		    	(void)strcpy( diagnostics1cp, "The Directory entry that this entry is to" ) ;
		    	(void)strcpy( diagnostics2cp, "update has not been specified" ) ;
		    	break ;

		    } /* else perhaps another search directory ? */
		}
		else
		{
	    	    /* Add the algoName and bulk object class attributes if required.. */
		    if( bulkLoadConfig.disableBulkAttributeUsage == FALSE )
		    	addAlgoNameAndBulkObjCls( readListp, resolvedEntryDN, sourceNamecp,
			    		purportedRDN, rootedAtSubDirListDetailsp -> subDirListRDN ) ;

	    	    freeListpp( &readListp ) ;
		    break ;
		}
	    }
	    else
	    {
	    	/* Creating a Directory entry.. */

                resolvedEntryDN = dn_cpy( resolvedParentDN ) ;
	    	dn_append( resolvedEntryDN, dn_comp_new( rdn_cpy( purportedRDN ) ) ) ;
		break ;
	    }


	    if( resolvedParentDN != NULLDN )
	    {
	    	dn_free( resolvedParentDN ) ;
                resolvedParentDN = NULLDN ;
	    }
	}

	if( resolvedParentDN != NULLDN )
	{
	    dn_free( resolvedParentDN ) ;
            resolvedParentDN = NULLDN ;
	}

	return( resolvedEntryDN ) ;
}



/*
 * For attempts to match an existing entry (i.e. purported entry name should already
 * have been created)..
 *
 * Note: an exact match and an algorithmic name match will have already been undertaken (and failed!)
 *
 * Test Case 1: For an RDN with "cn=P Barker" request the user to select one of the matches to
 *		a match for "cn=P*Barker"
 *
 * This procedure should only be run in interactive mode.
 */

int
unambiguousExistingEntryInteractiveDNresolution( searchListp, parentDN, purportedEntryDN, purportedRDN, resolvedDNp, sourceNamecp,
			rootedAtSubDirListDetailsp, diagnostics1cp, diagnostics2cp )
struct	namelist * searchListp ; /* algoName and objectClass attributes from the entries matching *INITIAL* SURNAME* */
DN	parentDN, purportedEntryDN ;
RDN	purportedRDN ;
DN   * 	resolvedDNp ;
char *	sourceNamecp ;
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
char *	diagnostics1cp, * diagnostics2cp ;
{
int 	ret_val = NOTOK, search_selected ;
struct	namelist * lp ;
char	selectionMsg[ SMALLBUFLEN ], * DNcp ;

	DNcp = dn2pstr( purportedEntryDN ) ;

	sprintf( selectionMsg, "For the entry name\n%s\nwhich of the suggested entry matches is correct", DNcp ) ;

	free( DNcp ) ;
	
	/* Run through each matched entry.. */

	lp = searchListp ;
	
	search_selected = enterSearchSelection( selectionMsg, searchListp, "Enter an alternative entry name" ) ;
		
	if( search_selected > NO_SELECTION )
	{
	    for( --search_selected ; search_selected > 0 ; search_selected-- )
		lp = lp -> next ;

	    /* Buffer the resolved DN.. */
	    if( (*resolvedDNp = dn_cpy( lp->entryDN )) == NULLDN )
	    {
		(void)strcpy( diagnostics1cp, "Internal error whilst unambiguously matching" ) ;
		(void)strcpy( diagnostics2cp, lp->name ) ;
	    }
	    else
	    {
	    	/* Add the algoName and bulk object class attributes if required.. */

		if( bulkLoadConfig.disableBulkAttributeUsage == FALSE )
		    addAlgoNameAndBulkObjCls( lp, *resolvedDNp, sourceNamecp,
		    	purportedRDN, rootedAtSubDirListDetailsp -> subDirListRDN ) ;

		ret_val = OK ;
	    }
	}
	else if( search_selected == EXTRA_SELECTION )
	{
	    if( (*resolvedDNp = userInputEntryOrParentDNresolution( parentDN, purportedEntryDN,
			rootedAtSubDirListDetailsp, purportedRDN, sourceNamecp,
				FALSE, FALSE, diagnostics1cp, diagnostics2cp )) != NULLDN )
	    {
	    	/* Add the algoName and bulk object class attributes if required.. */

		if( bulkLoadConfig.disableBulkAttributeUsage == FALSE )
		    addAlgoNameAndBulkObjCls( lp, *resolvedDNp, sourceNamecp,
		    	purportedRDN, rootedAtSubDirListDetailsp -> subDirListRDN ) ;

	    	ret_val = OK ;
	    }
	}
	else if( search_selected == NONE_OF_THE_SELECTION )
	    (void)strcpy( diagnostics1cp, "You have rejected all the possible existing entry matches" ) ;
	else
	    (void)strcpy( diagnostics1cp, "Internal error in unamibiguously match due to operator abort upon selection" ) ;
	
	return( ret_val ) ;
}



/*
 * If a read entry request finds an existing entry with the same RDN as an entry to 
 * be updated (as opposed to created) check if the commonName and personalTitle attribute
 * values permit the entries to be paired off.
 *
 * Note: readListp storage returned if match rejected
 */

#define MAXNEWTITLETYPES 3
#define MAXCOMMONNAMES	 3

int
RDNmatchMayBeAcceptedForEntryUpdate( currRAAp, updateRDN, readListp )
struct	rootedAtAttributes *  currRAAp ;
RDN	updateRDN ;
struct	namelist * readListp ; /* MUST include any cn and personalTitle attributes */
{
int	exactRDNmatchAccepted = TRUE, possTitlesMatch = TRUE, possForeNamesMatch = TRUE ;
struct  attributeLine * currALp = NULLALP ;
static	AttributeType cnat = NULLAttrT, ptat = NULLAttrT ;
struct	nameComponents	newNameStruct,      * newNameStructp      = &newNameStruct,
			existingNameStruct, * existingNameStructp = &existingNameStruct,
			newNameStructs[ MAXCOMMONNAMES ] ;
char *  cncp, * ptcp ;
int	newTitleTypes[ MAXNEWTITLETYPES ], maxUsedNewTitleTypesIndex = -1, maxUsedCNnameStructIndex = -1, i;
Attr_Sequence	as  ;
AV_Sequence	avs ;
RDN	ptrdncomp ;

	if( cnat == NULLAttrT )
	    cnat = str2AttrT( BL_COMMON_NAME ) ;

	/* Check if the RDN contains a commonName component.. */
	if( findAttrTypeRdnComp( cnat, updateRDN ) != NULLRDN )
	{
	    if( ptat == NULLAttrT )
		ptat = str2AttrT( BL_PERSONAL_TITLE ) ;

	    /* Set up the two name structures */
	    initialiseNameComponents( newNameStructp ) ;
	    initialiseNameComponents( existingNameStructp ) ;

	    /* Check for a personalTitle within the RDN.. */
	    if( (ptrdncomp = findAttrTypeRdnComp( ptat, updateRDN )) != NULLRDN )
	    {
	    	if( (ptcp = AttrV2pstr( &(ptrdncomp -> rdn_av)) ) != NULLCP )
		{
		    if( matchPersonalTitle( ptcp, newNameStructp ) == TRUE )
	    	    	newTitleTypes[ ++maxUsedNewTitleTypesIndex ] =
						newNameStructp -> titleType ;

		    free( ptcp ) ;
		}
	    }

	    /* Check for personal titles within the new attributes.. */
	    for( currALp = currRAAp->attrLinep ; currALp != NULLALP &&
	    		(maxUsedNewTitleTypesIndex+1) < MAXNEWTITLETYPES ; currALp = currALp -> nextAttrLinep )
	    {
	    	if( AttrT_cmp( ptat, currALp -> at ) == 0 )
		{
	    	    if( (ptcp = AttrV2pstr( currALp -> av )) != NULLCP )
		    {
		    	if( matchPersonalTitle( ptcp, newNameStructp ) == TRUE )
	    	    	    newTitleTypes[ ++maxUsedNewTitleTypesIndex ] =
						newNameStructp -> titleType ;

		    	free( ptcp ) ;
		    }
		}
	    }


	    /* Confirm that a personalTitle attribute is available for tests.. */
	    if( maxUsedNewTitleTypesIndex > -1 ) {

		/* Run through the existing entry's attribute types.. */
	    	for( as = readListp->ats ; possTitlesMatch == TRUE && as != NULLATTR ; as = as->attr_link )
	    	{
		    if( AttrT_cmp( as->attr_type, ptat ) == 0 )
		    {
		    	/* Run through the attribute values for this attribute type.. */
		    	for( avs = as->attr_value ; possTitlesMatch == TRUE && avs != NULLAV ; avs = avs->avseq_next )
		    	{
			    if( (ptcp = AttrV2pstr( &(avs->avseq_av) )) != NULLCP )
			    {
			    	if( matchPersonalTitle( ptcp, existingNameStructp ) == TRUE )
			    	{
			    	    for( i=0 ; possTitlesMatch == TRUE && i<=maxUsedNewTitleTypesIndex ; i++ )
                            	    	possTitlesMatch = possibleTitleTypesMatch( newTitleTypes[ i ],
			    						existingNameStructp->titleType ) ;
			    	}

			    	free( ptcp ) ;
			    }
		     	}
		    }
	    	}
	    }

            freeNameComponents( existingNameStructp ) ;
            freeNameComponents( newNameStructp ) ;


	    if( possTitlesMatch == FALSE )
	    	exactRDNmatchAccepted = FALSE ;
	    else
	    {
	    	/* Now prepare to check that the forenames presented for either the new or
	   	   existing entries do not cause us to reject the RDN match
	    	   e.g. if we have RDNs of "cn=P Bloggs" but one entry has
	    	   an alternative commonName of Peter Bloggs and the other Paul Bloggs. */

	    	for( i=0 ; i<MAXCOMMONNAMES ; i++ )
                    initialiseNameComponents( &(newNameStructs[ i ]) ) ;

	    	/* Check for common names within the new attributes.. */
	   	for( currALp = currRAAp->attrLinep ; currALp != NULLALP &&
	    	     (maxUsedCNnameStructIndex+1) < MAXCOMMONNAMES ; currALp = currALp -> nextAttrLinep )
	    	{
	    	    if( AttrT_cmp( cnat, currALp -> at ) == 0 )
		    {
	    	    	if( (cncp = AttrV2pstr( currALp -> av )) != NULLCP )
		    	{
	    	    	    extractNameComponents( cncp, NULLCP,
			    		&(newNameStructs[ ++maxUsedCNnameStructIndex ]) );

		    	    free( cncp ) ;
			}
		    }
		}

	    	/* Run through the existing entry's attribute types.. */
	    	for( as = readListp->ats ; possForeNamesMatch == TRUE && as != NULLATTR ; as = as->attr_link )
	    	{
		    if( AttrT_cmp( as->attr_type, cnat ) == 0 )
		    {
		    	/* Run through the attribute values for this attribute type.. */
		    	for( avs = as->attr_value ; possForeNamesMatch == TRUE && avs != NULLAV ; avs = avs->avseq_next )
		    	{
			    if( (cncp = AttrV2pstr( &(avs->avseq_av) )) != NULLCP )
			    {
			    	extractNameComponents( cncp, NULLCP, existingNameStructp );

			    	for( i=0 ; possForeNamesMatch == TRUE && i<=maxUsedCNnameStructIndex ; i++ )
				    possForeNamesMatch = possibleForenameMatch( existingNameStructp, &(newNameStructs[ i ])) ;

            			freeNameComponents( existingNameStructp ) ;
			       	free( cncp ) ;
			    }
			}
		    }
		}

	    	if( possForeNamesMatch == FALSE )
                    exactRDNmatchAccepted = FALSE ;

	    	for( i=0 ; i<MAXCOMMONNAMES ; i++ )
                    freeNameComponents( &(newNameStructs[ i ]) );
	    }
	}

	if( exactRDNmatchAccepted == FALSE )
	    freeListpp( &readListp ) ;

	return( exactRDNmatchAccepted ) ;
}


/*
 * Perform 1) an algorithmic name search
 *	   2) an exact RDN read entry retrieval
 *	   3) for personal entries perform a common name search
 *
 * ...as required to determine either an existing entry to be updated or
 * to ensure that when creating a new entry that no existing entries have a
 * name clash.
 */

int
determineActualDirectoryEntry( inFileType, operationalMode, OFD, sourceNamecp, currRAApp, prevRAAp,
				resolvedEntryDNp, actionReqp, rootedAtSubDirListDetailsp, updatesToExistingEntriesp )
int	inFileType, operationalMode ;
struct	outputFileDetails OFD[] ;
char *	sourceNamecp ;
struct	rootedAtAttributes ** currRAApp, /* May adjust this RDN block's attributes by an edit */
	* prevRAAp ;
DN   *	resolvedEntryDNp ;
int  *	actionReqp ; /* ADDACTIONREQ, DELETEACTIONREQ, NOACTIONREQ, MODRDNACTIONREQ */
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
int  *	updatesToExistingEntriesp ;
{
int	ret_val = OK, res, overruled = FALSE, overrulePossible = TRUE ;
struct	namelist *searchListp = NULLLIST, *algoNameAttrToBeReturnedp = NULLLIST,
	         *attrObjClsAlgNameCNtitlep = NULLLIST ;
struct	nameComponents entryNameStruct, * entryNameStructp = & entryNameStruct ;
char *	updateRDNcp, diagnostics1[ ERRORBUFSIZE ], diagnostics2[ ERRORBUFSIZE ] ;
RDN	updateRDN ;
DN	updateEntryDN, subDirListParentDN = NULLDN ;
int	DNloopExit = FALSE, isEntryToBeCreated, subDirListStatus = OK, firstIteration = TRUE, RDNreadResult = FALSE ;
struct  s_filter * uniqueKeyFilterp ;
searchAlgEnum	searchAlg ; /* Indicate which search algorithm (algoname or unique key)
			       succeeded in finding the directory entry */
int	valUKeyOK ;

	*resolvedEntryDNp = NULLDN ; /* returned on error */

	addToList( &algoNameAttrToBeReturnedp, BL_ALGO_NAME ) ;

	initialiseNameComponents( entryNameStructp ) ;

	uniqueKeyFilterp = (struct s_filter*)NULL ;


	while( DNloopExit == FALSE )
	{   
	    diagnostics1[0] = '\0', diagnostics2[0] = '\0' ;

	    isEntryToBeCreated = (*actionReqp == ADDRDNACTION || *actionReqp == MODRDNACTION ) ? TRUE : FALSE ;

	    if( isEntryToBeCreated == FALSE )
	    	overruled = FALSE ;

	    overrulePossible = FALSE ;

	    searchAlg = unknownSearchAlg ; valUKeyOK = OK ;

	    if( *actionReqp == MODRDNACTION )
	    {
	    	updateRDN   = rdn_cpy(  (*currRAApp) -> modRdn ) ;
	    	updateRDNcp = rdn2pstr( (*currRAApp) -> modRdn ) ;
	    }
	    else
	    {
	    	updateRDN   = rdn_cpy(  (*currRAApp) -> rdn ) ;
		updateRDNcp = rdn2pstr( (*currRAApp) -> rdn ) ;
	    }

       	    searchListp = NULLLIST ;

	    if( updateRDN != NULLRDN && updateRDNcp != NULLCP )
	    {
	    	/* Check if the initial "rootedAt" DN references a sub directory list
		   (on error the diagnostics will be set up) ... */

                if( (firstIteration == TRUE || subDirListStatus == NOTOK) &&
		    (subDirListStatus = checkForAndExpandSubDirList( rootedAtSubDirListDetailsp, (*currRAApp)-> rootedAtDN, diagnostics1 )) == NOTOK )
		{
		    /* A sub directory list was present but could not be expanded */
		}
		else if( bulkLoadConfig.disableBulkAttributeUsage == TRUE ||

			 (defineAndActionAlgoNameAttrSearch( sourceNamecp, (*currRAApp) -> rootedAtDN, updateRDN,
			    algoNameAttrToBeReturnedp, FALSE, &searchListp, rootedAtSubDirListDetailsp, &searchAlg, diagnostics1 ) == SEARCHOK && searchListp != NULLLIST) ||

			 ((valUKeyOK = validateUniqueKeySearchAndFormFilter( (*currRAApp)->attrLinep, &uniqueKeyFilterp )) == OK &&
			     defineAndActionUniqueKeySearch( (*currRAApp) -> rootedAtDN, &searchListp,
				rootedAtSubDirListDetailsp, uniqueKeyFilterp, &searchAlg, diagnostics1 ) == SEARCHOK) ||

			 (valUKeyOK == NOTOK) /* to ensure that the "validateUniqueKeySearchAndFormFilter" 
						 call does not prevent entry into the following block - i.e.
					         only search failures should skip the block*/ )
	    	{
		    /* ...Check to see if an "algoName" entry is already in existence
		       for this RDN at the specified rootedAt DN/sub tree list ... */

	    	    if( searchListp != NULLLIST )
		    {
		        /* An algoname or unique key match has indeed been found */
#ifdef BLT_DEBUG
fprintf( stderr, "\nSearch found %s", searchListp -> name ) ;
#endif
		    	if( isEntryToBeCreated == FALSE )
		    	{
		    	    *resolvedEntryDNp = str2dn( searchListp -> name ) ;

			    if( searchAlg == uniqueKeySearchAlg ) {
				addAlgoNameAndBulkObjCls( searchListp, *resolvedEntryDNp, sourceNamecp, 
						updateRDN, rootedAtSubDirListDetailsp->subDirListRDN ) ;
			    }

			    ret_val = OK ;
		    	}
		    	else
		    	{
			    if( *actionReqp != MODRDNACTION &&
			    	(bulkLoadConfig.autoAddEntryToAddAttrIfExists == TRUE ||
				 *updatesToExistingEntriesp		      == TRUE   ))
			    {
			    	/* Cause the change from entry creation to attribute addition */
				*actionReqp = NOACTIONREQ ;

		    	        *resolvedEntryDNp = str2dn( searchListp -> name ) ;

				if( searchAlg == uniqueKeySearchAlg ) {
				    addAlgoNameAndBulkObjCls( searchListp, *resolvedEntryDNp, sourceNamecp,
					updateRDN, rootedAtSubDirListDetailsp->subDirListRDN ) ;
				}

				ret_val = OK ;
			    }
			    else
			    {
			    	/* Entry already exists !! */
			    	(void)strcpy( diagnostics1, "Entry with this entry's required algorithmic name already exists for this source" ) ;
			    	(void)strcpy( diagnostics2, searchListp -> name ) ;
			    }
		    	}
		    
		    	freeListpp( &searchListp ) ;    

		    } else {
#ifdef BLT_DEBUG
fprintf( stderr, "\nAlgoName Search failed\n" );
#endif
			searchListp = NULLLIST ;

			updateEntryDN = dn_cpy( (*currRAApp) -> rootedAtDN ) ;

			dn_append( updateEntryDN, dn_comp_new( rdn_cpy( updateRDN ) ) ) ;
			
			addToList( &attrObjClsAlgNameCNtitlep , BL_OBJECT_CLASS ) ;
			addToList( &attrObjClsAlgNameCNtitlep , BL_ALGO_NAME ) ;
			addToList( &attrObjClsAlgNameCNtitlep , BL_COMMON_NAME ) ;
			addToList( &attrObjClsAlgNameCNtitlep , BL_PERSONAL_TITLE ) ;

		    	/*
		    	 * Try an exact RDN match..
		    	 */

			RDNreadResult = NOTOK ;

		    	if( rootedAtSubDirListDetailsp -> subDirListUsed == SUBDIRLISTNOTUSED &&
			    (RDNreadResult = readEntry( updateEntryDN, &searchListp, attrObjClsAlgNameCNtitlep, FALSE )) == OK &&
			    ( (isEntryToBeCreated == TRUE &&
			        (*actionReqp == MODRDNACTION ||
				    (bulkLoadConfig.autoAddEntryToAddAttrIfExists == FALSE &&
				     *updatesToExistingEntriesp			  == FALSE   ))) ||
			      RDNmatchMayBeAcceptedForEntryUpdate( *currRAApp, updateRDN, searchListp ) == TRUE) )
		        {
			    if( isEntryToBeCreated == FALSE )
		    	    {
				/* Update to existing entry.. */

			    	if( bulkLoadConfig.disableBulkAttributeUsage == FALSE ) {
			    	    /* Add the algoname to this entry.. */
			    	    addAlgoNameAndBulkObjCls( searchListp, updateEntryDN, sourceNamecp, updateRDN, NULLRDN );
			    	}

			    	*resolvedEntryDNp = str2dn( searchListp -> name ) ;
				ret_val = OK ;
			    }
			    else
			    {
			    	if( *actionReqp != MODRDNACTION &&
				    (bulkLoadConfig.autoAddEntryToAddAttrIfExists == TRUE ||
				     *updatesToExistingEntriesp      		  == TRUE   ))
			    	{
				    /* Update to existing entry.. */

			    	    if( bulkLoadConfig.disableBulkAttributeUsage == FALSE )
				    {
			    		/* Add the algoname to this entry.. */
			    		addAlgoNameAndBulkObjCls( searchListp, updateEntryDN, sourceNamecp, updateRDN, NULLRDN );
			    	    }

			    	    /* Cause the change from entry creation to attribute addition */
				    *actionReqp = NOACTIONREQ ;

			    	    *resolvedEntryDNp = str2dn( searchListp -> name ) ;
				    ret_val = OK ;
			    	}
			    	else
			    	{
			            /* Create entry request when entry RDN already exists !! */
			    
			            (void)strcpy( diagnostics1, "Entry with specified RDN already exists" ) ;
				    (void)strcpy( diagnostics2, searchListp -> name ) ;
			    	}
			    }
		    	}
		    	else
		    	{
			    /*
			     * Try for an unambiguous match of commonName
			     * and personalTitle attributes..
			     */

			    if( defineAndActionCnAndTitleSearch( (*currRAApp) -> rootedAtDN, 
			    		( *actionReqp == MODRDNACTION ) ? (*currRAApp)->modRdn : (*currRAApp)->rdn,
					(*currRAApp) -> attrLinep, entryNameStructp, &searchListp,
					rootedAtSubDirListDetailsp, diagnostics1 ) == SEARCHOK &&
				searchListp != NULLLIST )
	    		    {
			    	/* Check for a non-ambiguous name match between the directory
			           entries retrieved and the source data's RDN */

				res = NOTOK ;

				if( *updatesToExistingEntriesp == TRUE && *actionReqp != MODRDNACTION )
				{
				    res = unambiguousUpdateDNresolution( searchListp, entryNameStructp,
				    		sourceNamecp, resolvedEntryDNp, updateRDN,
				    			rootedAtSubDirListDetailsp -> subDirListRDN, OFD ) ;
							/* OFD added by J.C. 20/04/94 */

				    if( res == OK && isEntryToBeCreated == TRUE ) {

				    	/* Cause the change from entry creation to attribute addition */
					*actionReqp = NOACTIONREQ ; isEntryToBeCreated = FALSE ;
				    }
				}

				if( res == OK ) {
				    ret_val = OK ;
				} else {
	
				    if( isEntryToBeCreated == TRUE ) {

					if( rootedAtSubDirListDetailsp -> subDirListUsed != SUBDIRLISTNOTUSED ) {

				    	    if( operationalMode == INTERACT_OPER_MODE ) {

					    	if( subDirListParentDN != NULLDN ||
					    	    (subDirListParentDN = userInputEntryOrParentDNresolution( (*currRAApp)->rootedAtDN, updateEntryDN,
							rootedAtSubDirListDetailsp, updateRDN, sourceNamecp,
							TRUE, TRUE, diagnostics1, diagnostics2 )) != NULLDN )
					    	{
						    dn_free( updateEntryDN ) ;
						    updateEntryDN = dn_cpy( subDirListParentDN  ) ;
						    dn_append( updateEntryDN, dn_comp_new( rdn_cpy( updateRDN ) ) ) ;

					    	    /* Filter the previous results list */
						    filterSubTreeNamelistForSingleSubDirectory( subDirListParentDN, &searchListp ) ;
						}
					    }
					    else
					    {
					    	(void)strcpy( diagnostics1, "Sub-directory list reference in the \"rootedAt\". To create an entry" ) ;
					    	(void)strcpy( diagnostics2, "interactive input is needed to specify which sub-directory to use" ) ;
					    }
					}

					if( diagnostics1[ 0 ] == '\0' )
					{
                                            if( RDNreadResult != OK && (bulkLoadConfig.ignoreCloseNameMatches == TRUE ||
                                                overruled == TRUE) )
					    	res = OK ;
					    else 
					    	res = unambiguousCreateEntryDNresolution( *currRAApp, searchListp, 
					    		entryNameStructp, (*actionReqp == MODRDNACTION), diagnostics1, diagnostics2 ) ;

					    if( res == OK )
					    {
					    	/* .. no name clash with an existing entry */
					    
				   	    	*resolvedEntryDNp = dn_cpy( updateEntryDN ) ;
					    	ret_val = OK ;
					    }
					    else if( RDNreadResult != OK )
						overrulePossible = TRUE ;
					}
				    }
				    else
				    {
				    	if( operationalMode == INTERACT_OPER_MODE )
					    ret_val = unambiguousExistingEntryInteractiveDNresolution( searchListp, 
					    	(*currRAApp)->rootedAtDN, updateEntryDN, updateRDN, resolvedEntryDNp, sourceNamecp,
					    	rootedAtSubDirListDetailsp, diagnostics1, diagnostics2 ) ;

					else (void)strcpy( diagnostics1, "The RDN can only be matched against an existing entry in Interactive Mode" ) ;
				    }
				}
				
				freeListpp( &searchListp ) ;
			    }
			    else
			    {
#ifdef BLT_DEBUG
fprintf( stderr, "CN and titles Search failed to return anything \n" ) ;
#endif
				/* Could not find an existing entry that matches
				   the details - fine if we are creating it now !! */

				if( isEntryToBeCreated == FALSE )
				{
				    /* Entry update */

				    if( operationalMode == INTERACT_OPER_MODE )
				    {
	    			   	if( (*resolvedEntryDNp = userInputEntryOrParentDNresolution( (*currRAApp)->rootedAtDN, updateEntryDN,
							rootedAtSubDirListDetailsp, updateRDN, sourceNamecp,
							FALSE, FALSE, diagnostics1, diagnostics2 )) != NULLDN )
					{
				    	    ret_val = OK ;
					}
				    }
				    else
	    			    {
					(void)strcpy( diagnostics1, "Could not find the existing entry to update" ) ;
				    }
				}
				else
				{
				    if( rootedAtSubDirListDetailsp -> subDirListUsed == SUBDIRLISTNOTUSED )
				    {
				    	*resolvedEntryDNp = dn_cpy( updateEntryDN ) ;
					ret_val = OK ;
				    }
				    else
				    {
					/* We need to know which specific sub directory list sub-directory the entry is to be created in .. */
				    	if( operationalMode == INTERACT_OPER_MODE )
					{
	    			   	    if( (*resolvedEntryDNp = userInputEntryOrParentDNresolution( (*currRAApp)->rootedAtDN, updateEntryDN,
							rootedAtSubDirListDetailsp, updateRDN, sourceNamecp,
							FALSE, TRUE, diagnostics1, diagnostics2 )) != NULLDN )
					    {
				    	    	ret_val = OK ;
					    }
					}
					else
					{
					    (void)strcpy( diagnostics1, "Sub-directory list reference in the \"rootedAt\". To create an entry" ) ;
					    (void)strcpy( diagnostics2, "interactive input is needed to specify which sub-directory to use" ) ;
					}
				    }
				}
			    }

			    freeNameComponents( entryNameStructp ) ;
		    	}

			freeListpp( &attrObjClsAlgNameCNtitlep ) ;

		    	dn_free( updateEntryDN );
		    }
	    	}
#ifdef BLT_DEBUG
else fprintf( stderr, "Define and action failure \n" ) ;
#endif
		/* Check if the DN has been resolved.. */

		if( *resolvedEntryDNp != NULLDN )
		{
		    DNloopExit = TRUE ;
		}
		else if( operationalMode == NON_INTERACT_OPER_MODE )
		{
		    if( *actionReqp == MODRDNACTION )
			transferRootedAtModRDN( OFD, OUTPUTPROBLEMFILE, *currRAApp, diagnostics1, diagnostics2 ) ;
		    else
			transferRootedAtRDNattrs( OFD, OUTPUTPROBLEMFILE, *currRAApp, diagnostics1, diagnostics2 ) ;

	    	    DNloopExit = TRUE ;
	    	}
	    	else {

		    /* Interactive mode of operation.. */

		    DNloopExit = editRAandRDNblock( inFileType, OFD, sourceNamecp, currRAApp, prevRAAp, actionReqp, (*actionReqp == ADDRDNACTION) ? TRUE : FALSE, overrulePossible, 
		    			(overrulePossible == TRUE) ? "name clash" : NULLCP, &overruled, updatesToExistingEntriesp, diagnostics1, diagnostics2 ) ;

		    /* Note: the actionRequired may have been changed if an ADD RDN was changed
		       to a NO ACTION if the entry already exists but the values listed are still
		       to be added */
		}

		rdn_free( updateRDN ) ;
	    	free( updateRDNcp ) ;
	    }

	    firstIteration = FALSE ;

	} /* while loop */

	if( subDirListParentDN != NULLDN )
	    dn_free( subDirListParentDN ) ;
	
	if( searchListp != NULLLIST ) ;
	    freeListpp( &searchListp ) ;    

	freeListpp( &algoNameAttrToBeReturnedp ) ;


	return( ret_val ) ;
}
