/* bulkload.c */


/*****************************************************************************

  bulkload.c


12-APR-94	processInputFile( operational_mode, file_buffer, FALSE ) ;
		final parameter added - reported by John Colville.


*****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stdtypes.h>
#include <time.h>
#include <sys/time.h>

#include <isode/quipu/ufn.h>
#include <isode/quipu/util.h>
#include <isode/tailor.h>

#include "bulkload.h"
#include "bulkman.h"
#include "bind.h"
#include "namelist.h"
#include "bulkloadglobal.h"
#include "bulkloaderrors.h"
#include "bulkloadattrnames.h"
#include "configuration.h"
#include "readsearch.h"
#include "util.h"

LLog    _bulkload_log = {
    "bulkload.log", NULLCP, NULLCP, LLOG_FATAL | LLOG_EXCEPTIONS | LLOG_NOTICE,
            LLOG_NONE, -1, LLOGCLS | LLOGCRT | LLOGZER, NOTOK
};
LLog *bulkload_log = &_bulkload_log;
int bulkloadLogLevel = 1;


int boundToDSA = FALSE;
int limitProblem = FALSE;
int browseMess = FALSE;
int notAllReached = FALSE;
int alarmCount = 0;
int abandoned = FALSE;
int accessrightproblem = FALSE;
int searchfail = FALSE;
int controlCtoQuit = TRUE;
int wanAccess = FALSE;
int exactMatch;
int highNumber = 0;
int localAlarmTime = 15;
int remoteAlarmTime = 30;
int bindTimeout = 120;
int quitChars = 1;
int alarmLeft = 0;
	

/* Defaults for unspecified name parts */
char origDefaultCo[SMALLBUFLEN], origDefaultOrg[SMALLBUFLEN],
     exactString[SMALLBUFLEN],
     welcomeMessage[SMALLBUFLEN] = "Welcome to the Directory Service",
     matched[SMALLBUFLEN],
     callingDteNumber[SMALLBUFLEN];

char * username, * backup_dsa_address, * password;


extern  struct	bulkLoadConfigStruct	bulkLoadConfig ;
extern  struct	managementConfigStruct	managementConfig ;

void 	de_exit();

SFD	cleanupok();

void 	enterAndProcessManagementMenuOption() ;

char  *	bulkOIDs[] = { BL_DATA_SOURCE,
		       BL_MASTERS,
		       BL_CREATE_SOURCE,
		       BL_DELETE_SOURCE,
		       BL_ALGO_NAME,
		       BL_DELETE_ENTRY_AFTER,
		       BL_NON_MASTER_UPDATED_ATTR_TYPE,
		       BL_MASTER_UPDATED_ATTR_TYPE,
		       BL_VALIDATION_CHECK_AT,
		       BL_BULK_ENTRY_CLASS,
		       BL_BULK_SUB_DIR_LIST,
		       BL_BULK_SUB_TREE_DN,
		       BL_BULK_SUB_TREE_DN_EXT,
		       BL_ORGANIZATION,
		       BL_ORGANIZATIONAL_UNIT,
		       BL_LOCALITY,
		       BL_INHERITED,  /* Added by J.C. 20/01/94 */
		       NULLCP
		     } ;

int
checkOIDtables()
{
int	bulkOIDsChecked = OK ;
char **	bulkOIDscpp	= bulkOIDs ;
AttributeType at ;

	while( *bulkOIDscpp != NULLCP )
	{
#ifdef BLT_DEBUG
fprintf( stderr, "Checking for OID %s\n", *bulkOIDscpp ) ;
#endif	
	    if( (at = str2AttrT( *bulkOIDscpp )) == NULLAttrT )
	    {
	    	fprintf( stderr, "Initialisation test failed, %s not in \"oidtable.at\"\n", *bulkOIDscpp ) ;
		bulkOIDsChecked = NOTOK ;

	    } else {
		AttrT_free( at ) ; /* DUMMY */
	    }

	    bulkOIDscpp++ ;
	}

	return( bulkOIDsChecked ) ;
}


/*
 * main : check for and request the required parameters before
 *        initiating file processing.
 */

int main(argc, argv)
     int argc;
     char *argv[];
{
int	operational_mode = INVALID_OPER_MODE, managementMode = FALSE ;
char	file_buffer[SMALLBUFLEN], *cp ;

	file_buffer[0] = '\0' ;

	if (initialisations( argc, argv, &operational_mode, file_buffer, &managementMode ) != OK)
	    de_exit(-1);

	if( checkOIDtables() != OK )
	    de_exit( ERRCHECKOIDFAILURE ) ;

	if (de_bind(FALSE) != OK)
	    de_exit( ERRBINDFAILURE );

	ll_hdinit(bulkload_log, "bulkload");

	if( managementMode == FALSE )
	{
	    welcome() ; pageprint( "\n\n" ) ;
	}

	(void) signal(SIGINT, cleanupok);


	if (boundToDSA == FALSE)
      	    if (de_bind(TRUE) == NOTOK) /* block */
       		de_exit( ERRBINDFAILURE );
	
	pagerOn( NUMBER_NOT_ALLOWED );

	if( managementMode == FALSE )
	{
	    /* Check if the operational mode has been entered.. */

	    if( operational_mode == INVALID_OPER_MODE )
	    {
	    	if( enter_oper_mode( &operational_mode ) == NOTOK )
	    	    de_exit( ERRNOOPERMODEENTERED ) ;
	    }


	    /* Check if the file name has been entered.. */
	    if( file_buffer[0] == '\0' )
	    {
	    	if( enter_file_name( file_buffer ) == NOTOK )
		    de_exit( ERRNOFILENAMEENTERED ) ;
	    }

#ifdef BLT_DEBUG
fprintf( stderr, "oper mode %d file name [%s]\n", operational_mode, file_buffer );
#endif

	    if( operational_mode == INTERACT_OPER_MODE )
	    {
	        if( (cp = getenv( "EDITOR" )) == NULLCP || (*cp == '\0') )
	        {
	    	    fprintf( stderr, "Initialisation failure, your $EDITOR environment variable is missing\n" ) ;
		
		    de_exit( ERREDITORNOTDEFINED ) ;
	    
	    	} else if( bulkLoadConfig.displayConfigAtStart == TRUE ) {

	    	    /* Indicate the current configuration details to the user (allowing updates).. */
	    	    processConfigBulkLoadUpdates() ;
	    	}
	    }

	    processInputFile( operational_mode, file_buffer, FALSE ) ;
            /* 3rd argument was omitted from distribution version
               Change made by J.C. 14/01/94 */

	} else { 

	    /* Management operations are required */
	    if( (cp = getenv( "EDITOR" )) == NULLCP || (*cp == '\0') )
	    {
		fprintf( stderr, "Initialisation failure, your $EDITOR environment variable is missing\n" ) ;
		de_exit( ERREDITORNOTDEFINED ) ;
	    }

	    enterAndProcessManagementMenuOption() ;
	}

	de_exit( 0 ) ;

	return( 0 ) ;
}

int
enter_file_name( file_bufferp )
char *	file_bufferp ;
{
int 	ret_val = OK ;

	resetprint( "Which file name is to be processed" );
	
	pageprint( "\n:- " ) ;

	if( fgets( file_bufferp, SMALLBUFLEN, stdin ) == NULLCP ||
	    		strlen( file_bufferp ) == 0 ||
			lexnequ( file_bufferp, "quit", strlen( file_bufferp )) == 0 )
	    ret_val = NOTOK ;
	else
	{
	    /* Remove the trailing whitespace */
	    *(file_bufferp + strlen( file_bufferp ) - 1 ) = '\0' ;
	}

	pageprint( "\n" ) ;

	return( ret_val ) ;
}

int
enter_oper_mode( oper_modep )
int *	oper_modep ;
{
char	line_buffer[SMALLBUFLEN] ;
int	ret_val = NOTOK ;

	resetprint( "Do you require the " );
	writeInverse( "I" ) ;
	pageprint( "nteractive or " );
	writeInverse( "N" );
	pageprint( "on-interactive operational mode" );
	

	for( ; ; )
	{
	    pageprint( "\n(I/N)[I]:- " ) ;

	    if( fgets( line_buffer, SMALLBUFLEN, stdin ) == NULLCP ||
		   	strlen( line_buffer ) == 0 ||
			tolower( line_buffer[0] ) == 'q' )
		break ;

	    if( isspace( line_buffer[0] ) )
	    {
	    	/* Default required */
	    	*oper_modep = INTERACT_OPER_MODE ;
		ret_val = OK ;
		break ;
	    }
	    else if( tolower( line_buffer[0] ) == NON_INTERACTIVE_CHAR )
	    {
		*oper_modep = NON_INTERACT_OPER_MODE ;
		ret_val = OK ;
		break ;
	    }
	    else if( tolower( line_buffer[0] ) == INTERACTIVE_CHAR )
	    {
	    	*oper_modep = INTERACT_OPER_MODE ;
		ret_val = OK ;
		break ;
	    }
	}

	pageprint( "\n" ) ;	

	return( ret_val ) ;
}

SFD 
cleanupok()
{
	/* Ensure that any unprocessed lines are passed to the PROBLEMs file. */
	(void)processInputFileRootedAts( -1, -1, NULL, NULL, NULL, NULLCP, NULLDN, FALSE, TRUE, ERRABANDONED ) ;

	fprintf( stderr, "FATAL error : Session abandoned\n" ) ;

	de_exit( ERRABANDONED ) ;
}

void
handleTimeout()
{
	(void) signal(SIGALRM, SIG_IGN);

	/* Ensure that any unprocessed lines are passed to the PROBLEMs file. */
	(void)processInputFileRootedAts( -1, -1, NULL, NULL, NULL, NULLCP, NULLDN, FALSE, TRUE, ERRTIMEOUT ) ;

	fprintf( stderr, "FATAL error : DSA timed out\n" ) ;

	de_exit( ERRTIMEOUT ) ;
}


void
de_exit( exitCode )
int	exitCode;
{
	if (boundToDSA == TRUE)
      	    (void) de_unbind();

	exit(exitCode);
}


/*
 * Defaults to NO...
 */
char	enterYesNo( str1, str2 )
char * 	str1, * str2 ;
{
char	buf[SMALLBUFLEN];

  	for( ; ; )
  	{
	    resetprint( "" ) ;
	    
	    (void) pageprint( "%s\n%s(y/n)[n]? ", str1, str2);

	    if (gets(buf) == NULLCP) /* control-D */
		return 'n';

    	    pageprint( "\n" );

	    buf[0] = tolower( buf[0] ) ;

    	    if( buf[0] == 'y' || buf[0] == 'n' )
      		return buf[0];

	    else if( buf[0] == '\0' )
		return 'n' ;
  	}
}


/*
 * Note: updating the configuration will not cause procedure exit until
 * a futher option has been selected.
 */

int
enterEditResponse( convertAddEntryToAddEntrysValuesReq, editOfRDNBlock, overridePossible, overridecp,
							disenableMinusEforEntryOptionReq )
int	convertAddEntryToAddEntrysValuesReq, editOfRDNBlock, overridePossible ;
char *	overridecp ;
int	disenableMinusEforEntryOptionReq ;
{
char	line_buffer[SMALLBUFLEN], promptBuffer[SMALLBUFLEN] ;
int	editResponse = EDITNOTOK ;

	resetprint( "" ) ;
	writeInverse( "P" );
	pageprint( "rocess edit updates,\n" ) ;

	if( disenableMinusEforEntryOptionReq == TRUE )
	{
	    pageprint( "disenable the updates to e" ) ;
	    writeInverse( "X" ) ;
	    pageprint( "isting entry option for this entry,\n" ) ;
	}

	if( overridePossible == TRUE && overridecp != NULLCP )
	{
	    pageprint( "overri" ) ;
	    writeInverse( "D" ) ;
	    pageprint( "e %s\n", overridecp ) ;
	}
	
	if( convertAddEntryToAddEntrysValuesReq == TRUE )
	{	
	    pageprint( "convert the add entry to an add of the entry's " ) ;
	    writeInverse( "V" ) ;
	    pageprint( "alues,\n" ) ;
	}

	if( editOfRDNBlock == TRUE )
	{
	    pageprint( "route the original attributes " ) ;
	    writeInverse( "I" ) ;
	    pageprint( "ndividually\n" ) ;
	}

	writeInverse( "R" ) ;
	pageprint( "e-edit the file,\n" ) ;

	pageprint( "pass " ) ;
	writeInverse( "O" ) ;
	pageprint( "riginal to problems file,\n" ) ;

	pageprint( "pass " ) ;
	writeInverse( "E" ) ;
	pageprint( "dit to problems file,\n" ) ;

	pageprint( "pass " ) ;
	writeInverse( "O" ) ;
	pageprint( "riginal to " ) ;
	writeInverse( "N" ) ;
	pageprint( "o actions file,\n" ) ;

	pageprint( "pass " ) ;
	writeInverse( "E" ) ;
	pageprint( "dit to " ) ;
	writeInverse( "N" ) ;
	pageprint( "o actions file or\n" ) ;

	writeInverse( "**" ) ;
	pageprint( " " ) ;
	writeInverse( "C" ) ;
	pageprint( "onfiguration updates " ) ;
	writeInverse( "**\n" ) ;
	
	writeInverse( "**" ) ;
	pageprint( "   " ) ;
	writeInverse( "H" ) ;
	pageprint( "elp information    " ) ;
	writeInverse( "**" ) ;
	
	for( ; ; )
	{
	    sprintf( promptBuffer, "%s%s%s%s%s%s", "\n(P",
		(disenableMinusEforEntryOptionReq == TRUE) ? "/X" : "",
	        (overridePossible == TRUE && overridecp != NULLCP) ? "/D" : "",
	    	(convertAddEntryToAddEntrysValuesReq == TRUE) ? "/V" : "", 
	    	(editOfRDNBlock == TRUE) ? "/I" : "", 
		"/R/O/E/ON/EN/C/H)[P]:- " ) ;

	    pageprint( promptBuffer ) ;
	    
	    if( fgets( line_buffer, SMALLBUFLEN, stdin ) == NULLCP ||
		   	strlen( line_buffer ) == 0 ||
			tolower( line_buffer[0] ) == 'q' )
		break ;

	    if( isspace( line_buffer[0] ) ||
	        lexnequ( line_buffer, EDIT_PROCESS_EDIT_STR, 2 ) == 0 )
	    {
	    	/* Default required */
		editResponse = EDITEDITREPROCESS ;
		break ;
	    }
	    else if( disenableMinusEforEntryOptionReq == TRUE &&
	    	     lexnequ( line_buffer, EDIT_DISENABLE_MINUS_E_STR, 2 ) == 0 )
	    {
	    	editResponse = EDITDISENABLEMINUSEFORENTRY ;
		break ;
	    }
	    else if( overridePossible == TRUE && overridecp != NULLCP &&
	    	     lexnequ( line_buffer, EDIT_OVERRIDE, 2 ) == 0 )
	    {
	    	editResponse = EDITOVERRIDE ;
		break ;
	    }
	    else if( convertAddEntryToAddEntrysValuesReq == TRUE &&
	    	     lexnequ( line_buffer, EDIT_ADD_ENT_TO_ADD_VALS_STR, 2 ) == 0 )
	    {
	     	editResponse = EDITADDENTRYCONVTOADDVALUES ;
		break ;
	    }
	    else if( editOfRDNBlock == TRUE &&
	    	     lexnequ( line_buffer, EDIT_ROUTE_INDIV_ATTRS_STR, 2 ) == 0 )
	    {
	    	editResponse = EDITROUTEINDIVATTRS ;
	    	break ;
	    }
	    else if( lexnequ( line_buffer, EDIT_REEDIT_FILE_STR, 2 ) == 0 )
	    {
	     	editResponse = EDITREEDITFILE ;
		break ;
	    }
	    else if( lexnequ( line_buffer, EDIT_SEND_ORIG_PROB_STR, 2 ) == 0 )
	    {
	     	editResponse = EDITSENDORIGTOPROBLEMSFILE ;
		break ;
	    }
	    else if( lexnequ( line_buffer, EDIT_SEND_EDIT_PROB_STR, 2 ) == 0 )
	    {
	     	editResponse = EDITSENDEDITTOPROBLEMSFILE ;
		break ;
	    }
	    else if( lexnequ( line_buffer, EDIT_SEND_ORIG_NOACT_STR, 2 ) == 0 )
	    {
	     	editResponse = EDITSENDORIGTONOACTFILE ;
		break ;
	    }
	    else if( lexnequ( line_buffer, EDIT_SEND_EDIT_NOACT_STR, 2 ) == 0 )
	    {
	     	editResponse = EDITSENDEDITTONOACTFILE ;
		break ;
	    }
	    else if( lexnequ( line_buffer, EDIT_CONFIG_FILE_STR, 2 ) == 0 )
	    {
	     	/* editResponse does not need to be set up, but for completeness.. **/
		editResponse = EDITCONFIGFILE ;

	    	processConfigBulkLoadUpdates() ;
	    }
	    else if( lexnequ( line_buffer, EDIT_HELP_INFO_STR, 2 ) == 0 )
	    {
	     	/* editResponse does not need to be set up, but for completeness.. **/
		editResponse = EDITHELPINFORMATION ;

	    	editFile( isodefile( "bulk/bulkloadhelp", 0 ) ) ;
	    }
	}
#ifdef BLT_DEBUG
fprintf( stderr, "Edit response %d\n", editResponse ) ;
#endif
	resetprint( "\n" ) ;	

	return( editResponse ) ;
}


int  
enterSearchSelection( msgcp, listp, furtherOptioncp )
char *	msgcp ;
struct	namelist * listp ;
char *	furtherOptioncp  ;
{
int	selection=0, maxSelection=0 ;
struct	namelist * lp = listp ;

	resetprint( "\n" );

	pageprint( "%s\n", msgcp );

	/* Display the list of possibilities */

    	while( (lp != NULLLIST) && (selection <= 0) )
    	{
	    maxSelection++ ;
	 
	    pageprint( "%d)%s\n", maxSelection, lp->name );

	    lp = lp -> next ;
    
	    /* Retrieve the selection entered at the pager prompt */
    	    selection = getpnum() ;

	} /* while loop */

    	if( selection <= 0 )
    	{
    	    /* Allow for the possibility that it is none of these.. */
    	    pageprint( "%d) NONE OF THE ABOVE\n", ++maxSelection );
    
	    if( furtherOptioncp != NULLCP )
	    	pageprint( "%c) %s\n", *furtherOptioncp, furtherOptioncp ) ;

	    selection = enterSelection( 1, maxSelection, furtherOptioncp ) ;
    	}

    	return( selection );
}


int
enterSelection( minOption, maxOption, furtherOptioncp )
int	minOption, maxOption ;
char *	furtherOptioncp ;
{
char 	buf[ SMALLBUFLEN ], *cp ;
int  	ret_val ;

  	for( ; ; )
  	{
	    if( furtherOptioncp != NULLCP )
    		pageprint( "\n(%d..%d,%c) ", minOption, maxOption, *furtherOptioncp ) ;
            else
    		pageprint( "\n(%d..%d) ", minOption, maxOption ) ;

	    if (gets(buf) == NULLCP) /* control-D */
		return NO_SELECTION ;

    	    cp = buf ;

    	    /* Skip leading spaces */
    	    while( isspace( *cp ) )
		cp++ ;

    	    if( !cp )
    		continue ;

	    if( furtherOptioncp != NULLCP && tolower( *cp ) == tolower( *furtherOptioncp ) )
	    	return( EXTRA_SELECTION ) ;
    
    	    ret_val = atoi( cp ) ;

    	    if(( ret_val >= minOption) && (ret_val <= maxOption))
    	    {
      		if( ret_val == maxOption )
      		    ret_val = NONE_OF_THE_SELECTION ;
	
		pageprint( "\n" );

      		return( ret_val ) ;
    	    }
  	}
}


DN
enterUpdateEntryRDNandFormDN( parentDN, purportedEntryDN, readListpp )
DN	parentDN, purportedEntryDN ;
struct	namelist ** readListpp ;
{
char 	buf[ SMALLBUFLEN ], *cp, *DNcp ;
DN	resolvedDN = NULLDN, enteredDN ;
RDN	enteredRDN ;

	DNcp = dn2pstr( purportedEntryDN ) ;

	for( ; ; )
	{
	    resetprint( "\nSource entry name =\n%s\n", DNcp ) ;
	    resetprint( "Enter the replacement entry name component e.g.\"cn=Paul Barker\" (*<return> to return)\n" ) ;
	    
	    if (gets(buf) == NULLCP) /* control-D */
		break ;

    	    cp = buf ;

    	    /* Skip leading spaces */
    	    while( isspace( *cp ) )
		cp++ ;

	    if( *cp == '*' )
	    	break ;

	    if( (enteredRDN = str2bulkRdn( cp )) == NULLRDN )
	    {	    	
	    	resetprint( "INVALID name component entered\n" ) ;
		continue ;
	    }
	    else
	    {
		enteredDN = dn_cpy( parentDN ) ;

		dn_append( enteredDN, dn_comp_new( enteredRDN ) ) ;

		/* Return all attributes if entry exists.. */
	  	if( readEntry( enteredDN, readListpp, NULLLIST, TRUE ) == OK && *readListpp != NULLLIST )
		{
		    resolvedDN = enteredDN ;
		    /* Note: readListpp storage returned by a higher level procedure */
		    break ;
		}
		else 
		{
		    dn_free( enteredDN ) ;
		    resetprint( "The name component does not exist under the parent entry\n" ) ;
		}
	    }
	}

	free( DNcp ) ;

	return( resolvedDN ) ; /* If NULLDN then "*" or CTRL-D exit */
}


void
enterAndProcessManagementMenuOption()
{
char 	buf[ SMALLBUFLEN ], *cp;
int	menuOption, minOption = MENUMINOPTION, maxOption = MENUMAXOPTION ;

	if( bulkLoadConfig.disableBulkAttributeUsage == TRUE )
	{
	    minOption = maxOption = MENUOPTDUMPSUBTREE ;
	}
  	
	for( ; ; )
  	{
	    resetprint( "" ) ;

	    if( bulkLoadConfig.disableBulkAttributeUsage == FALSE )
	    {
	    	pageprint( "\n%d) Report on master/create and delete sources", MENUOPTSRCEREPORT ) ;
	    	pageprint( "\n%d) Check for expired entries", MENUOPTEXPIREDENTRIES ) ;
	    }
	    
	    pageprint( "\n%d) Dump Directory Subtree", MENUOPTDUMPSUBTREE ) ;

	    if( bulkLoadConfig.disableBulkAttributeUsage == FALSE )
	    {
	    	pageprint( "\n%d) Reset \"validationCheckAt\" within a Subtree", MENUOPTRESETVALIDCHKAT ) ;
	    	pageprint( "\n%d) Compare a data source to the Directory", MENUOPTCHKSRCAGNSTDIR ) ;
	    	pageprint( "\n%d) Check for entries that have not be compared against", MENUOPTCHKVALIDCHKAT ) ;
	    }
	    
	    pageprint( "\n%d) Perform Integrity check on attributes with the DN syntax", MENUOPTCHKDNS ) ;
	    pageprint( "\n%d) Prepare for the insertion of \"seeAlso\" cross references to \"roleOccupant\" attributes", MENUOPTXREFROLOCCSEEALSO ) ;
	

/*	    pageprint( "\n THE FOLLOWING IS A TEMPORARY OPTION FOR THE CHANGE OVER" ) ;
	    pageprint( "\n BETWEEN masterUpdatedAttrType AND nonMasterUpdatedAttrType" ) ;
	    pageprint( "\n%d) Exchange or Remove the masterUpdatedAttrTypes", MENUEXCHANGEMASTERUPAT ) ;*/

/*	    pageprint( "\n%d) Add value to subtree nodes", MENUOPTADDVALTOSUBTREENODES ) ; */

	    pageprint( "\nc) **Update the management configuration**" ) ;
	    pageprint( "\nh) **Help information**" ) ;
    	    pageprint( "\nq) QUIT" ) ;
	    pageprint( "\n(%d..%d/c/h/q) ", minOption, maxOption ) ;

	    if (gets(buf) == NULLCP) /* control-D */
		break ;

    	    cp = buf ;

    	    /* Skip leading spaces */
    	    while( isspace( *cp ) )
		cp++ ;

    	    if( !cp )
    		continue ;

	    if( tolower( *cp ) == 'q' )
	    	break ;

	    if( tolower( *cp ) == 'h' )
	    {
	    	editFile( isodefile( "bulk/bulkmanhelp", 0 ) ) ;	    
	    	continue ;
	    }
	    
	    if( tolower( *cp ) == 'c' )
	    {
	    	processConfigManagementUpdates() ;
	    	continue ;
	    }
    
    	    menuOption = atoi( cp ) ;

    	    if( menuOption >= minOption && menuOption <= maxOption )
    	    {
		pageprint( "\n" );

		switch( menuOption )
		{
		    case MENUOPTSRCEREPORT :

			processSourceReport( FALSE ) ;

			break ;

		    case MENUOPTEXPIREDENTRIES :

			processExpiredEntries( FALSE ) ;
		    	
			break ;
      		
		    case MENUOPTDUMPSUBTREE :

			processDumpSubtree( FALSE ) ;
		        
			break ;

		    case MENUOPTADDVALTOSUBTREENODES :

		    	processAddValueToSubTreeNodes( FALSE ) ;

			break ;
		    
		    case MENUOPTRESETVALIDCHKAT :

			processResetValidCheckAt( FALSE ) ;

		    	break ;

		    case MENUOPTCHKSRCAGNSTDIR :

		    	processSourceValidation( FALSE ) ;

		    	break ;

		    case MENUEXCHANGEMASTERUPAT :

			/* ..option disabled */
		    	processExchangeMasterUpAtforNonMasterUpAt( FALSE ) ;

		    	break ;

		    case MENUOPTCHKVALIDCHKAT :
			/* ..option disabled */
		    	processCheckValidCheckAt( FALSE ) ;

		    	break ;

		    case MENUOPTCHKDNS :
			processCheckDNIntegrity( FALSE ) ;

			break ;

		    case MENUOPTXREFROLOCCSEEALSO :
			processCrossRefRoleOccAndSeeAlso( FALSE ) ;

			break ;
		}
    	    }
  	}
	
	return ;
}
