#include "ctype.h"
#include "string.h"


#include <isode/quipu/util.h>
#include <isode/quipu/attrvalue.h>

#include "util.h"
#include "bulkloadglobal.h"
#include "Bulk-types.h"


PE
AlgoName_enc( parm )
struct	algoName_struct * parm ;
{
PE 	ret_pe = NULLPE ;

	if( parm != NULLALGONAMESTRUCT )
		encode_Bulk_AlgoNameSyntax(&ret_pe, 0, 0, NULLCP, parm) ;

	return( ret_pe ) ;
}

struct	algoName_struct *
AlgoName_dec( pe )
PE	pe ;
{
struct	algoName_struct * result = NULLALGONAMESTRUCT ;

	if( pe != NULLPE )
		if( decode_Bulk_AlgoNameSyntax(pe, 0, NULLIP, NULLVP, &result) == NOTOK )
			return( NULLALGONAMESTRUCT ) ;

	return( result ) ;
}


struct	algoName_struct *
AlgoName_cpy( parm )
struct	algoName_struct * parm ;
{
struct	algoName_struct * algo_structp = NULLALGONAMESTRUCT ;

	if( (algo_structp = (struct algoName_struct *) calloc(1, sizeof(struct algoName_struct))) != NULLALGONAMESTRUCT )
	{
		if( parm -> sourceNamecp != NULLCP )
			algo_structp -> sourceNamecp = strdup( parm -> sourceNamecp ) ;
		else
			algo_structp -> sourceNamecp = NULLCP ;

		if( parm -> sourceEntryNameRDN != NULLRDN )
			algo_structp -> sourceEntryNameRDN = rdn_cpy( parm -> sourceEntryNameRDN ) ;
		else
			algo_structp -> sourceEntryNameRDN = NULLRDN ;
	}
	
	return( algo_structp ) ;
}


/* Input string of the form "ccData#cn=Paul Barker%personalTitle=Mr" */

struct	algoName_struct *
str2AlgoName( cp )
char *	cp ;
{
struct	algoName_struct * algo_structp = NULLALGONAMESTRUCT ;
char *	buffcp = NULLCP, * hashcp ;
char *	endOfSourceNamecp, * startOfRDNcp ;


	if( (cp != NULLCP) && 
	    ((buffcp = strdup( cp )) != NULLCP ) &&
	    ((hashcp = strchr( buffcp, '#' )) != NULLCP) &&
	    (hashcp > buffcp + 1 ) &&
	    (algo_structp = (struct algoName_struct *) calloc(1, sizeof(struct algoName_struct))) != NULLALGONAMESTRUCT )
	{
		endOfSourceNamecp = hashcp ;

		*endOfSourceNamecp = '\0' ; /* ..Split off the source name */

		removeAllWhiteSpaces( buffcp ) ;

		algo_structp -> sourceNamecp = strdup( buffcp ) ;


		startOfRDNcp = hashcp + 1 ;

		algo_structp -> sourceEntryNameRDN = NULLRDN ;

		if( *startOfRDNcp )
		{
			while( isspace( *startOfRDNcp ) )
				startOfRDNcp++ ; /* ..Skip leading space */

			algo_structp -> sourceEntryNameRDN = str2rdn( startOfRDNcp ) ;
		}
	}

	if( buffcp != NULLCP )
		free( buffcp ) ;
	
	return( algo_structp ) ;
}


void
AlgoName_print( ps, parm, format )
PS	ps ;
struct	algoName_struct * parm ;
int	format ;
{
	if( parm != NULLALGONAMESTRUCT )
	{
		ps_printf( ps, "%s", parm -> sourceNamecp ) ;
		ps_print(  ps, " # " ) ;
		rdn_print( ps, parm -> sourceEntryNameRDN, format ) ;
		ps_print(  ps, " " ) ;
	}

	return ;
}


char * AlgoName2pstr() ;

int
AlgoName_cmp( a, b )
struct	algoName_struct * a, *b ;
{
int	ret = 0 ;

	if( a == NULLALGONAMESTRUCT && b == NULLALGONAMESTRUCT )
		ret = 0 ;

	else if( a == NULLALGONAMESTRUCT )
		ret = -1 ;

	else if( b == NULLALGONAMESTRUCT )
		ret = 1 ;


	else if( a -> sourceNamecp == NULLCP && b -> sourceNamecp == NULLCP )
		ret = 0 ;

	else if( a -> sourceNamecp == NULLCP )
		ret = -1 ;
	
	else if( b -> sourceNamecp == NULLCP )
		ret = 1 ;


	else {
		ret = strcasecmp( a->sourceNamecp, b->sourceNamecp ) ;

		if( ret == 0 )
		{
			if( a -> sourceEntryNameRDN == NULLRDN && b -> sourceEntryNameRDN == NULLRDN )
				ret = 0 ;

			else if( a -> sourceEntryNameRDN == NULLRDN )
				ret = -1 ;

			else if( b -> sourceEntryNameRDN == NULLRDN )
				ret = 1 ;

			else ret = rdn_cmp( a->sourceEntryNameRDN, b->sourceEntryNameRDN ) ;
		}
	}

	/* Due to the usage of strcasecmp.. */
	if( ret > 0 )
		ret =  1 ;
	else if( ret < 0 )
		ret = -1 ;

	return( ret ) ;
}

/*
 * Approximate match solely for the data source component..
 */
 
int
AlgoName_approx( a, b )
struct	algoName_struct * a, *b ;
{
int	ret = 0 ;

	if( a == NULLALGONAMESTRUCT && b == NULLALGONAMESTRUCT )
		ret = 0 ;

	else if( a == NULLALGONAMESTRUCT )
		ret = -1 ;

	else if( b == NULLALGONAMESTRUCT )
		ret =  1 ;


	else if( a -> sourceNamecp == NULLCP && b -> sourceNamecp == NULLCP )
		ret = 0 ;

	else if( a -> sourceNamecp == NULLCP )
		ret = -1 ;
	
	else if( b -> sourceNamecp == NULLCP )
		ret = 1 ;


	else ret = strcasecmp( a->sourceNamecp, b->sourceNamecp ) ;

	/* Due to the usage of strcasecmp.. */
	if( ret > 0 )
		ret =  1 ;
	else if( ret < 0 )
		ret = -1 ;

	return( ret ) ;
}


void
AlgoName_free( parm )
struct	algoName_struct * parm ;
{
	if( parm != NULLALGONAMESTRUCT )
	{
		if( parm -> sourceNamecp != NULLCP )
			(void)free( parm -> sourceNamecp ) ;

		if( parm -> sourceEntryNameRDN != NULLRDN )
			rdn_free( parm -> sourceEntryNameRDN ) ;

		(void)free( (char *) parm ) ;
	}

	return ;
}

  
void 
AlgoNameSyntaxHandler()
{
	add_attribute_syntax( "AlgoNameSyntax", (IFP) AlgoName_enc, (IFP) AlgoName_dec,
				(IFP) str2AlgoName, (IFP)AlgoName_print, (IFP)AlgoName_cpy,
				(IFP) AlgoName_cmp, (IFP)AlgoName_free, NULLCP, (IFP) AlgoName_approx, TRUE ) ;
	return ;
}


static PS psalg = NULLPS;

char *
AlgoName2pstr( algp )
struct	algoName_struct * algp ;
{
    char       *cp;

    if( (psalg == NULL
	    && ((psalg = ps_alloc (str_open)) == NULLPS))
		    || str_setup (psalg, NULLCP, SMALLBUFLEN, 0) == NOTOK) {
	if (psalg)
	    ps_free (psalg), psalg = NULLPS;

	return NULLCP;
    }

    AlgoName_print( psalg, algp, EDBOUT );

    *--psalg -> ps_ptr = NULL, psalg -> ps_cnt++;

    cp = psalg -> ps_base ;

    psalg -> ps_base = NULL, psalg -> ps_cnt = 0;
    psalg -> ps_ptr  = NULL, psalg -> ps_bufsiz = 0;

    return cp;
}


void
testAlgoNameSyntax()
{
struct	algoName_struct * algp, *algp1, *algp2, * algp3 ;
char *	cp, * sourceEntryNameRDNcp ;
PE	algpe ;

	algp  = str2AlgoName( "ccData # cn=Paul Barker%personalTitle=Mr" ) ;
	algp1 = str2AlgoName( "ccData # cn=Kevin McCarthy" );

	fprintf( stderr, "1) Compare result [1] %d\n", AlgoName_cmp( algp, algp1 ) ) ;

	
	sourceEntryNameRDNcp = rdn2pstr( algp->sourceEntryNameRDN ) ;

	fprintf( stderr, "Source Name %s\nEntry Name %s\n",
			algp->sourceNamecp, sourceEntryNameRDNcp ) ;

	free( sourceEntryNameRDNcp ) ;


	algpe = AlgoName_enc( algp ) ;

	algp2 = AlgoName_dec( algpe ) ;

	pe_free( algpe ) ;


	algp3 = AlgoName_cpy( algp1 ) ;

	fprintf( stderr, "2) Compare result [1] %d\n", AlgoName_cmp( algp2, algp3 ) ) ;
	
	
	fprintf( stderr, "3) Compare result [0] %d\n", AlgoName_cmp( algp, algp2 ) ) ;

	cp = AlgoName2pstr( algp2 ) ;

	fprintf( stderr, "AlgoName %s\n", cp ) ;

	free( cp ) ;

	
	AlgoName_free( algp2 ), AlgoName_free( algp1 ), AlgoName_free( algp ) ; 


	return ;
}
